# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of popEye
#
# popEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# popEye is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cadenceEtl.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import  (absolute_import, division, print_function)
__metaclass__ = type

import re
import datetime

class fncBase ():
    def __init__(self,*args, **kargs):
        self.cDate = datetime.datetime.today().strftime('%m/%d/%y')

    def handler(self, col): pass

class fDCast(fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)

    def handler(self, col):
        if not col:
            return self.cDate
        if len(col)!=8 or not col.isdigit ():
            return None

        yy = col[:4]
        mm = col[4:6]
        dd = col[6:8]

        if yy=="0000" or yy=="9999" or len(yy.strip())<4:
            return None

        if yy<'1900' :
            return '01/01/2006'
        if yy>'2050' :
            yy = '2050'
        return None if yy == "9999" else "%s/%s/%s" % (mm, dd, yy)

class fDTCast(fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)

    def handler(self, col):
        hh = '00'
        min= '00'
        ss = '00'
        yy = "0000"
        mm = '00'
        dd = '00'

        if not col:
            return None

        if len (col)>=8 and col.isdigit ():
            yy = col[:4]
            mm = col[4:6]
            dd = col[6:8]

            if len(col) == 14 :
                hh = col[8:10]
                min = col[10:12]
                ss = col[12:14]

                if hh == "24":
                    hh = "00"

            if yy=="0000" or yy=="9999" or len(yy.strip())<4:
                return None

            if yy<'1900' :
                return '01/01/2006'

            if yy>'2050':
                yy = "2050"

        return None if yy == "0000" or yy=="9999" else "%s/%s/%s %s:%s:%s" % (mm, dd, yy, hh,min,ss)

class fDFile(fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)

    def handler(self, col):
        if not col or len(col)==0:
            return col
        col = col[:10].split("/")

        if len(col)==3:
            yy = col[2]
            mm = col[1]
            dd = col[0]

            return "%s/%s/%s" % (mm, dd, yy)
        col = col[:10].split("-")
        if len(col)==3:
            yy = col[0]
            mm = col[1]
            dd = col[1]
            return "%s/%s/%s" % (mm, dd, yy)

        return None

class fDCurr(fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)

    def handler(self, col):
         return self.cDate

class fTCast(fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)

    def handler(self, col):
        if not col or "000000" in col:
            return None
        hh = col[:2]
        mm = col[2:4]
        ss = col[4:6]
        return  "23:59:29" if "24" in hh else "%s:%s:%s" % (hh, mm, ss)

class fR(fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)
        self.rep0 = args[0]
        self.rep2 = args[1]

    def handler(self, col):
        if col and isinstance(col, str):
            return col.replace (self.rep0, self.rep2)
        return col

class fNull(fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)
        self.rep0 = args[0]

    def handler(self, col):
        if not col or len (col.strip()) == 0:
            return self.rep0
        return col

class fClob(fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)

    def handler(self, col):
        if not col:
            return None

        return col


#############################OLD  ATTEMT #############################################################

class fPhone (fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)

    def handler(self, col):
        phone = re.sub(r'\D', '', col)
        return phone

class fDecode (fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)
        self.xx = "Nothing"
        if len(args)>0: self.xx = args[0]

    def handler(self, col):
        return str(col)+" sss"

class fAddress (fncBase):
    def __init__(self, *args, **kargs):
        fncBase.__init__(self, *args, **kargs)
        self.adressType = ""
        if len(args)>0:
            self.adressType = args[0]

    def handler(self, col):
        return col+self.adressType






# -*- coding: utf-8 -*-
# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of popEye
#
# popEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# popEye is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cadenceEtl.  If not, see <http://www.gnu.org/licenses/>.

import time
import sys
import os
import traceback

import multiprocessing.pool as mpool
from collections import OrderedDict

from popEtl.config import config
from popEtl.glob.glob import p, setQueryWithParams, decodeStrPython2Or3
from popEtl.glob.loaderFunctions import *
from popEtl.glob.enums import eDbType, eConnValues, isDbType
import popEtl.connections.dbQueries as queries
import popEtl.connections.dbQueryParser as queryParser

# Data sources
aConnection = [x.lower() for x in config.CONNECTIONS_ACTIVE]
if eDbType.SQL in aConnection or eDbType.ACCESS in aConnection :
    import pyodbc as odbc #   ceODBC as odbc #pyodbc  # pyodbc version: 3.0.7

if eDbType.MYSQL  in aConnection   :
    import pymysql as pymysql

if eDbType.VERTIVA in aConnection :
    import vertica_python
    # pip install vertica_python
    # Need to install pip install sqlalchemy-vertica-python as well !!!

if eDbType.ORACLE in aConnection  :
    import cx_Oracle                      # version : 6.1

class cnDb (object):
    def __init__ (self, connDic=None, connType=None, connName=None, connUrl=None, connFilter=None):
        self.cIsSql     = connDic [ eConnValues.connIsSql] if connDic else None
        self.cName      = setQueryWithParams ( connDic [ eConnValues.connObj] ) if connDic else connName
        self.cSchema    = None
        self.cType      = connDic [ eConnValues.connType] if connDic else connType

        if self.cIsSql:
            self.cSQL = self.cName
        elif self.cName:
            tblName = self.__wrapSql(col=self.cName, remove=False)
            self.cSQL = "SELECT * FROM "+tblName

        self.cColumns   = []
        # Will be update if there is a query as source and mapping in query as well (select x as yy.....
        self.cColumnsTDic= None
        self.cUrl       = connDic [ eConnValues.connUrl ] if connDic else connUrl
        self.cWhere     = connDic [ eConnValues.connFilter ] if connDic else connFilter
        self.cColoumnAs = True

        self.insertSql  = None

        if not self.cType or not isDbType(self.cType):
            p ("Connection type is not valid: %s, use connection from config file" %(str(self.cType)) )
            return
        if  not self.cUrl:
            p("Connection URL is not exists, use valid URL conn" )
            return

        objName = "query" if self.cIsSql else self.cName
        p("db->init: DB type: %s, table: %s" % (self.cType, objName, ), "ii")

        if eDbType.MYSQL == self.cType:
            self.conn = pymysql.connect(self.cUrl["host"], self.cUrl["user"], self.cUrl["passwd"], self.cUrl["db"])
            self.cursor = self.conn.cursor()
        elif eDbType.VERTIVA == self.cType:
            self.conn = vertica_python.connect(self.cUrl)
            self.cursor = self.conn.cursor()
        elif eDbType.ORACLE == self.cType:
            self.conn = cx_Oracle.connect(self.cUrl['user'], self.cUrl['pass'], self.cUrl['dsn'])
            if 'nls' in self.cUrl:
                os.environ["NLS_LANG"] = self.cUrl['nls']
            self.cursor = self.conn.cursor()
        elif eDbType.ACCESS == self.cType:
            self.conn       = odbc.connect (self.cUrl) # , ansi=True
            self.cursor     = self.conn.cursor()
            self.cColoumnAs = False
        else:
            self.conn = odbc.connect (self.cUrl) #ansi=True
            self.cursor = self.conn.cursor()

        if not self.cIsSql and self.cName:
            self.cName = self.__wrapSql(col=self.cName, remove=True)
            self.cName = self.cName.split(".")

            self.cSchema =self.cName[0] if len(self.cName) > 1 else config.DATA_TYPE['schema'][self.cType]
            self.cName =  self.cName[1] if len(self.cName) > 1 else self.cName[0]

            if self.cWhere and len (self.cWhere)>1:
                self.cWhere = re.sub (r'WHERE', '', self.cWhere, flags=re.IGNORECASE)
                self.cWhere = setQueryWithParams (self.cWhere)
                self.cSQL = self.cSQL + " WHERE " + self.cWhere

    def close(self):
        try:
            if self.cursor: self.cursor.close()
            if self.conn:   self.conn.close()
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            p("sb->close: Exception :"+str(exc_type)+" file name:"+str(fname)+" line: "+str(exc_tb.tb_lineno)+" massage: "+str(exc_obj.message), "e")

    def setColumns(self, sttDic):
        ret = OrderedDict ()
        columnsList = [(i, sttDic[i]["t"]) for i in sttDic if "t" in sttDic[i]]

        dType = config.DATA_TYPE.keys()

        for tup in columnsList:
            fName = tup[0]
            fType = tup[1]

            fmatch = re.search ("\(.+\)",fType)

            if fmatch:
                fType = re.sub ("\(.+\)","",fType)

            if fType in dType:
                rType= config.DATA_TYPE[fType][self.cType]
                if (isinstance(rType, tuple )):
                    rType = rType[0]
            else:
                rType= config.DATA_TYPE['default'][self.cType]
                fmatch = None
            fullRType = str(rType)+""+str(fmatch.group()) if fmatch else rType

            self.cColumns.append ( (fName , fullRType.lower()) )
        p("db->setColumns: type: %s, table %s will be set with column: %s" %(self.cType,self.cName, str(self.cColumns) ), "ii")
        return

    def create(self, stt=None,  seq=None, tblName=None):
        tblName = tblName if tblName else "[" + self.cSchema + "].[" + self.cName+"]" if self.cSchema else "["+self.cName+"]"
        colList = [(t,stt[t]["t"]) for t in stt if "t" in stt[t]] if stt else self.getColumns()

        if colList and len (colList)>0:
            boolToCreate = self.__cloneObject(colList, tblName)

            col = "("
            # create new table
            if boolToCreate:
                sql = "CREATE TABLE "+tblName+" \n"
                if seq:
                    colSeqName = self._wrapSql(col=seq['column'], remove=True)
                    colType    = getattr(queries, self.cType + "_seq")(seq)
                    col += "["+colSeqName+"]"+"\t"+ colType
                for colTup in colList:
                    colName = colTup[0].replace("[","").replace("]","")
                    if colName != colTup[1]:
                        col += "["+colName+"]"+"\t"+ colTup[1] +",\n"
                col = col[:-2]
                col+=")"
                sql += col
                p ("Create table \n"+sql)
                self.__executeSQL (sql)
        else:
            p("db->create: Table %s cannot create - problem with mappong source columns, src column: %s " %( str(self.cName), str(colList) ), "e")

    def truncate(self, tbl=None):
        tbl = tbl if tbl else self.cName
        sql = getattr(queries, self.cType + "_truncate")(tbl)

        self.__executeSQL(sql)
        p("db->truncate: truncate table DB type: %s, table: %s, url: %s" % (self.cType, self.cName, str(self.cUrl)),"ii")

    def getColumns (self):
        if self.cColumns and len(self.cColumns)>0:
            return self.cColumns
        else:
            self.structure (stt=None)
        return self.cColumns

    def structure (self, stt, tableName=None, addSourceColumn=False, sqlQuery=None):
        tableStructure  = []
        # If there is query and there is internal maaping in query - will add this mapping to mappingColum dictionary
        if self.cIsSql:
            stt = self.__sqlQueryMapping (stt=stt, addSourceColumn=addSourceColumn, sqlQuery=sqlQuery)
            for t in stt:
                if "s" in  stt[t] and "t" in stt[t]:
                    tableStructure.append ( (stt[t]["s"] , stt[t]["t"]) )
        else:
            # Get closing and starting column
            sttTemp = OrderedDict()
            sttSource = {}
            #get all source from stt
            if stt:
                for t in stt:
                    if "s" in stt[t]:
                        if stt[t]["s"] not in sttSource:
                            sttSource[stt[t]["s"]] = t
                        else:
                            tmpList = sttSource[stt[t]["s"]] if isinstance(sttSource[stt[t]["s"]], list) else sttSource[stt[t]["s"]].split()
                            tmpList.append ( t )
                            sttSource[stt[t]["s"]] = tmpList


            tableName = tableName if tableName else self.cName
            if self.cType in ('access'):
                rows = self.__access (tableName)
            else:
                sql = getattr(queries, self.cType + "_columnDefinition")(tableName)
                self.__executeSQL(sql, commit=False)
                rows = self.cursor.fetchall()
            for row in rows:
                cName = row[0]
                cType = row[1].lower().replace(' ','')
                if addSourceColumn or stt is None:
                    tableStructure.append( ( cName,cType ) )
                else:
                    if unicode(cName) in sttSource:
                        tableStructure.append((cName, cType))
                if cName in sttSource:
                    targetKey  = sttSource[cName]
                    if isinstance(targetKey, list):
                        for tKey in targetKey:
                            val = stt[tKey]
                            if "t" not in val: val["t"] = cType
                            if tKey not in sttTemp: sttTemp[tKey] = val
                    else:
                        val = stt[targetKey]
                        if "t" not in val: val["t"] = cType
                        if targetKey not in sttTemp: sttTemp[targetKey] = val
                else:
                    targetKey   = cName
                    val         = {"s":cName,"t":cType}
                    if targetKey not in sttTemp:
                        if (stt is not None and addSourceColumn) or stt is None:
                            sttTemp[targetKey] = val

            if stt:
                for t in stt:
                    if t not in sttTemp: sttTemp[t] = stt[t]


            if len(sttTemp)>0: stt = sttTemp

        self.cColumns = tableStructure
        return stt

    def loadData(self, srcVsTar, results, numOfRows, cntColumn):
        if self.cIsSql:
            p("db->loadData: Object is query... connot load data, %s " %(self.cName))
            return

        if results and numOfRows>0:
            tarSQL = "INSERT INTO " + self.cName + " "
            if srcVsTar and len(srcVsTar)>0:
                tarL = [self.__wrapSql(col=t[1], remove=False, cType=self.cType) for t in srcVsTar]
                tarSQL += "(" + ','.join(tarL) + ") "
                tarSQL += "VALUES (" + ",".join(["?" for x in range(len(tarL))]) + ")"
            else:
                cntColumn = cntColumn if cntColumn else len (results[0])
                tarSQL += "VALUES (" + ",".join(["?" for x in range(cntColumn)]) + ")"

            try:
                self.cursor.executemany(tarSQL, results)
                self.conn.commit()
                p('db->loadData: Load %s into target: %s >>>>>> ' % (str(numOfRows), self.cName), "ii")
            except Exception as e:
                p("db->loadData: type: %s, name: %s ERROR in cursor.executemany !!!!" % (self.cType, str(self.cName)), "e")
                p("db->loadData: ERROR, target query: %s " % str(tarSQL), "e")
                p("db->loadData: ERROR, sample result: %s " % str(results[0]), "e")
                p(e, "e")

                if config.RESULT_LOOP_ON_ERROR:
                    p("db->loadData: ERROR, Loading row by row  ", "e")
                    iCnt = 0
                    tCnt = len (results)
                    for r in results:
                        try:
                            iCnt+=1
                            r = [r]
                            self.cursor.executemany(tarSQL, r)
                            self.conn.commit()
                        except Exception as e:
                            ret = ""
                            for col in r[0]:
                                if col is None:
                                    ret += str(col) + ","
                                elif str(col).replace(".", "").replace(",", "").isdigit():
                                    ret += str(col) + " ,"
                                else:
                                    ret += "'" + str(col) + "' ,"
                            p("db->loadData: ERROR, LOOPING ON ALL RESULTS, ROW ERROR ", "e")
                            p(e, "e")
                            p(tarSQL, "e")
                            p(ret, "e")
                    p("db->loadData: ERROR Row by row: loader %s out of %s  " %(str(iCnt),str(tCnt)) , "e")
        return

    def transferToTarget(self, dstObj, srcVsTar, fnDic, pp):
        srcSql = self.__dbMapSrcVsTarget(srcSql=self.cSQL, srcVsTar=srcVsTar)
        try:
            self.__executeSQL(str(srcSql), commit=False)
            columnsInSOurce     = [x[0] for x in self.cursor.description]
            totalColumnInSource = len(columnsInSOurce)

            p ('db->transferToTarget: Loading total columns:%s, object name: %s  ' %(str(totalColumnInSource), str(dstObj.cName)),"ii")
            self.__parallelProcessing (dstObj=dstObj, srcVsTar=srcVsTar, fnDic=fnDic, pp=pp,  cntColumn=totalColumnInSource)

        except Exception as e:
            p("db->transferToTarget: ERROR Exectuging query: %s, type: %s >>>>>>" % (srcSql, self.cType) , "e")
            p(str(e), "e")

    def __dbMapSrcVsTarget ( self,srcSql, srcVsTar ):
        # there is column mapping or function mapping
        if not srcVsTar or len(srcVsTar)==0:
            return srcSql

        stcSelect = srcSql.lower().replace("\n", " ").find("select ")
        stcFrom   = srcSql.lower().replace("\n", " ").find(" from ")

        if stcSelect > -1 and stcFrom > 0:
            preSrcSql = srcSql[:stcSelect + 7]
            postSrcSql= srcSql[stcFrom:]
            newCol    = ""
            for tup in srcVsTar:
                if self.cIsSql:
                    srcC = tup[0]
                else:
                    srcC = self.__wrapSql(col=tup[0], remove=False) if tup[0] != "''" else tup[0]

                srcT = self.__wrapSql(col=tup[1], remove=False)
                newCol += srcC + " AS " + srcT + "," if self.cColoumnAs else srcC + ","

            newCol = newCol[:-1]
            srcSql = preSrcSql + newCol + postSrcSql
            #p("db->__dbMapSrcVsTarget: there is mapping, update to new sql query: %s " % (srcSql), "ii")
        return srcSql

    def __parallelProcessing (self, dstObj, srcVsTar, fnDic, pp, cntColumn=None):
        numOfRows = 0
        iCnt      = 0
        pool = mpool.ThreadPool(config.NUM_OF_LOADING_THREAD)

        'An iterator that uses fetchmany to keep memory usage down'
        while True:
            try:
                if pp:
                    results = self.cursor.fetchmany(config.RESULT_ARRAY_SIZE)
                else:
                    results = self.cursor.fetchall()
                results = self.__functionResultMapping( results, fnDic)
            except Exception as e:
                p("db->__parallelProcessing: type: %s, name: %s ERROR in cursor.fetchmany" %(self.cType, str(self.cName)), "e")
                p(str(e), "e")
                break
            if not results or len(results)<1:
                break
            numOfRows+=len(results)
            pool.apply_async(func=self.__parallelProcessingLoad,args=(dstObj, srcVsTar, results, numOfRows, cntColumn))
            if iCnt < config.NUM_OF_LOADING_THREAD:
                iCnt+=1
            else:
                pool.close()
                pool.join()
                pool = mpool.ThreadPool(config.NUM_OF_LOADING_THREAD)
                iCnt = 0
        if pool:
            pool.close()
            pool.join()

    def __parallelProcessingLoad (self, dstObj, srcVsTar, results, numOfRows, cntColumn=None):
        try:
            return dstObj.loadData (srcVsTar, results, numOfRows, cntColumn)
        except Exception as e:
            p(e,'e')
            traceback.print_exc()
            raise e

    def __functionResultMapping (self, results,fnDic):
        if fnDic and len(fnDic) > 0 and results:
            for cntRows, r in enumerate(results):
                r = list(r)
                for pos, fnList in fnDic.items():
                    if not isinstance(pos, tuple):
                        uColumn = r[pos]
                        for f in fnList:
                            uColumn = f.handler(uColumn)
                        r[pos] = uColumn
                    else:
                        fnPos = fnList[0]
                        fnStr = fnList[1]
                        fnEval = fnList[2]
                        newVal = [str(r[cr]).decode(config.FILE_DECODING) for cr in pos]
                        newValStr = unicode(fnStr).format(*newVal)
                        r[fnPos] = eval(newValStr) if fnEval else newValStr
                results[cntRows] = r
        return results

    def minValues (self, colToFilter, resolution=None, periods=None, startDate=None):
        # there is min value to
        sql = getattr(queries, self.cType + "_minValue")(self.cName, self.cSchema, resolution, periods, colToFilter, startDate)
        p ("db->minValues: exec query : %s" %(sql), "ii")
        self.__executeSQL(sql)
        minValue = self.cursor.fetchone()
        if minValue and len (minValue)>0:
            minValue = minValue[0]
        else:
            p ("db->minValues: ERROR Getting miniumum value sql: "+sql, "e")
            return None
        p("db->minValues: get minimum value for table %s, field %s, sql : %s" %(str( self.cType), str(colToFilter), str(sql)), "ii" )
        return minValue

    def execSP (self, sqlQuery ):
        self.__executeSQL( sqlQuery )

    def merge (self, mergeTable, mergeKeys, sourceTable=None ):
        self.__sqlMerge(mergeTable, mergeKeys,sourceTable=sourceTable)

    def cntRows (self):
        sql = ""
        if self.cIsSql:
            sql = "SELECT COUNT (*) FROM ("+self.cName+")"
        else:
            tblName = self.cName.split(".")
            tblName = tblName[0] if len(tblName)==1 else tblName[1]
            sql = "SELECT COUNT (*) FROM ["+tblName+"]"
        self.__executeSQL(sql, commit=False)
        rows = self.cursor.fetchall()
        res = rows[0][0] if len(rows)>0 and len (rows[0])>0 else 0
        return res

    def select (self, sql):
        self.__executeSQL(sql=sql, commit=False)
        return self.cursor.fetchall()

    def getExistColumns (self, tblName=None):
        existStrucute = []
        tblName = tblName if tblName else self.cName
        tblName = self.__wrapSql(col=tblName, remove=True)
        objectExists = self.__objectExists(objName=tblName)
        if (objectExists):
            p("db-> __cloneObject: Table %s is exist >>>>" % (tblName), "ii")

            # get all current strucute of existing table
            sql = getattr(queries, self.cType + "_columnDefinition")(tblName)
            self.__executeSQL(sql, commit=False)

            rows = self.cursor.fetchall()
            for row in rows:
                colName = self.__wrapSql(col=row[0], remove=False)
                colType = row[1].lower().replace(' ', '')
                existStrucute.append((colName, colType))
        return existStrucute

    def __chunker(self, seq, size):
        return (seq[pos:pos + size] for pos in xrange(0, len(seq), size))

    def __cloneObject(self, colList, tblName=None):
        colList = [(str(self.__wrapSql(col=tup[0], remove=False) ),tup[1].lower().replace (" ","")) for tup in colList]
        tblName = tblName if tblName else  self.cName
        existStrucute = self.getExistColumns(tblName=tblName)

        if config.TABLE_HISTORY:
            p ("db-> __cloneObject: Table History is ON ...","ii")
            oldName     = None
            schemaEqual = True

            if len(existStrucute)>0:
                schemaEqual = True if existStrucute == colList  else False

                if not schemaEqual:
                    srcPost = config.DATA_TYPE['colFrame'][self.cType][1]
                    p("db-> __cloneObject: UPDATE TABLE OLD STRUCTURE : %s " % str(existStrucute))
                    p("db-> __cloneObject: OLD STRUCTURE : %s " %str(existStrucute))
                    p("db-> __cloneObject: NEW STRUCTURE : %s " % str(colList))

                    if tblName[-1]=="]":
                        oldName = "%s_%s]" %(tblName[:-1], str(time.strftime('%y%m%d')))
                    else:
                        oldName = "%s_%s" %(tblName, str (time.strftime('%y%m%d')))

                    if (self.__objectExists(objName=oldName)):
                        num = 0
                        while (self.__objectExists(objName=oldName)):
                            num += 1
                            if tblName[-1] == "]":
                                oldName = "%s_%s_%s]" % (tblName[:-1], str(time.strftime('%y%m%d')), str(num))
                            else:
                                oldName = "%s_%s_%s" % (tblName, str(time.strftime('%y%m%d')), str(num))
                            oldName = oldName[: oldName.rfind('_')] + "_" + str(num)
                    if oldName:
                        p ("db-> __cloneObject: Table History is ON and changed, table %s exists ... will rename to %s" %(str (tblName) , str(oldName) ), "ii")
                        oldName = oldName[oldName.find('.')+1:]
                        oldName = self.__wrapSql(col=oldName, remove=True)
                        tblName = self.__wrapSql(col=tblName, remove=True)
                        sql = getattr (queries, self.cType+"_renameTable")(tblName,oldName)

                        #sql = eval (self.objType+"_renameTable ("+self.objName+","+oldName+")")
                        p("db-> __cloneObject: rename table, sql: %s" % (str(sql)), "ii")
                        self.__executeSQL(sql)
                else:
                    p("db-> __cloneObject: No changes made in table %s >>>>>" % (tblName), "ii")
                    return False
        else:
            if len(existStrucute)>0:
                p("db-> __cloneObject: Table History is OFF, table exists, will drop table %s... " % (str(tblName)), "ii")
                sql = eval (self.cType+"_renameTable("+tblName+")")
                self.__executeSQL(sql)
                return True
            else:
                p("db-> __cloneObject: Table History is OFF, table not exists exists, will create table %s... " % (str(tblName)), "ii")
        return True

    def __objectExists (self, objName=None):
        objName = self.cName if not objName else objName
        sql = "Select OBJECT_ID('"+objName+"')"
        self.cursor.execute(sql)
        row = self.cursor.fetchone()
        if row[0]:
            p ("db-> __objectExists: table %s exists ..." %(str (objName)) , "ii")
            return True
        p ("db-> __objectExists: Table %s is not exists ..." %(str (objName)) , "ii")
        return False

    def __executeSQL(self, sql, commit=True):
        if not (isinstance(sql, (list,tuple))):
            sql = [sql]
        try:
            for s in sql:
                s = unicode(s)
                self.cursor.execute(s)  # if 'ceodbc' in odbc.__name__.lower() else self.conn.execute(s)
            if commit:
                self.conn.commit()          # if 'ceodbc' in odbc.__name__.lower() else self.cursor.commit()
            return True
        except Exception as e:
            if e.args:
                error = e.args
                msg = decodeStrPython2Or3 (error, un=False)
            else:
                msg = e
            p("db->__executeSQL: ERROR : ", "e")
            p(e, "e")
            p("db->__executeSQL: ERROR %s " % str(msg), "e")
            p("db->__executeSQL: ERROR SQL: %s " %(sql),"e" )
            return False

    def __schemaCompare (self, colList):
        if self.cColumns == colList:
            p('db-> __schemaCompare: table exists with same strucure as column list ... ', "ii")

            tableStructure = self.structure()
            p('db-> __schemaCompare: table %s, old structure: %s, new: %s' %( str(self.cName), str(tableStructure), str(self.cColumns)), "ii")
            if set (self.cColumns) == set (tableStructure):
                p('db-> __schemaCompare: table %s has no change >>>' %(self.cType), "ii")
                return True
        p ('db-> __schemaCompare: table %s structure changed, old: %s, new: %s >>>>' %(self.cName, str(self.cColumns), str(colList) ), "ii")
        return False

    def __sqlMerge(self, mergeTable, mergeKeys,sourceTable):
        dstTable = self.__wrapSql(col=mergeTable, remove=False)

        if sourceTable:
            columns = self.getExistColumns(tblName=sourceTable)
            srcCol = [c[0] for c in columns]
        else:
            srcTable = self.__wrapSql(col=self.cName, remove=False)
            if self.cSchema:
                srcTable = "%s.%s" % (self.__wrapSql(col=self.cSchema, remove=False), srcTable)

            self.cColumns = self.getColumns()
            srcCol = [c[0] for c in self.cColumns]

        trgCol = srcCol
        # test
        colList     = []
        colFullList = []
        colOnList   = []
        for c in srcCol:
            if c in trgCol:
                colFullList.append(c)
                if c not in mergeKeys:
                    colList.append(c)
                else:
                    colOnList.append(c)

        notValidColumn = list(set(mergeKeys)-set(colOnList))
        if notValidColumn and len (notValidColumn)>0:
            p ("db->__sqlMerge: Not valid column %s " %(str (notValidColumn) ) , "ii")

        if len (colOnList)<1:
            mergeKeys = colList

        # dstTable, srcTable, mergeKeys, colList , colFullList
        colList = [self.__wrapSql(col=x, remove=False)  for x in colList]
        colFullList = [self.__wrapSql(col=x, remove=False)  for x in colFullList]
        sql = getattr(queries, self.cType + "_merge")(dstTable, srcTable, mergeKeys, colList , colFullList)
        self.__executeSQL(sql)
        p("db->sqlServer_Merge: Merged source %s table with %s table as target" % (srcTable, dstTable), "ii")

    def __sqlQueryMappingHelp (self,allTableStrucure, col):
        boolFind    = False
        colType     = None
        colTbl      = None
        colName     = None
        #colList     = col.split(".")
        #if len(colList) == 2:
        #    tblName = colList[0]
        #    colNameL = colList[1].lower()
        #else:
        tblName = None
        colNameL = col.lower()

        if tblName and tblName in allTableStrucure:
            if colNameL in allTableStrucure[tblName]:
                boolFind    = True
                colType     = allTableStrucure[tblName][colNameL][2]
                colTbl      = allTableStrucure[tblName][colNameL][1]
                colName     = allTableStrucure[tblName][colNameL][0]
            else:
                columnsName = allTableStrucure[tblName].keys()
                for colOrg in columnsName:
                    if colOrg in colNameL:
                        boolFind = True
                        colType = allTableStrucure[tblName][colOrg][2]
                        colTbl = allTableStrucure[tblName][colOrg][1]
                        colName = colNameL
                        break

        elif not tblName:
            for tblName in allTableStrucure:
                if colNameL in allTableStrucure[tblName]:
                    boolFind = True
                    colType = allTableStrucure[tblName][colNameL][2]
                    colTbl  = allTableStrucure[tblName][colNameL][1]
                    colName = allTableStrucure[tblName][colNameL][0]
                    break
                else:
                    columnsName = allTableStrucure[tblName].keys()
                    for colOrg in columnsName:
                        if colOrg in colNameL:
                            boolFind = True
                            colType = allTableStrucure[tblName][colOrg][2]
                            colTbl  = allTableStrucure[tblName][colOrg][1]
                            colName = colNameL
                            break
        if not boolFind:
            p("db->_sqlQueryMappingHelp there is column mapping which is not exists in any source table, ignoring. column: %s, tables: %s " % (str(col), str(allTableStrucure.keys())), "ii")
        return colType, colTbl, colName

    def __sqlQueryMapping (self,stt=None, addSourceColumn=False, sqlQuery=None):
        tableStructure  = []
        mappingDic      = {}
        sqlQ            = sqlQuery if sqlQuery else self.cName if self.cIsSql else None
        sttTemp         = None

        if sqlQ and len(sqlQ)>0:
            sttTemp             = OrderedDict()
            # sqlQ                = sqlQ.replace ("'",'"')
            columnTblDic        = queryParser.extract_tableAndColumns(sqlQ)

            allColumnsList      = [x for x in columnTblDic[config.QUERY_ALL_COLUMNS_KEY]]
            allColumnsTarget    = [x for x in columnTblDic[config.QUERY_TARGET_COLUMNS]]
            alldistinctColumn   = []
            allTableStrucure    = {}

            # Update alldistinctColumn
            for col in allColumnsList:
                colSplit = col.split(".",1)

                alldistinctColumn.append ( ('',colSplit[0]) if len(colSplit)==1 else (colSplit[0],colSplit[1]) )

            # update allTableStrucure dictionary : {tblName:{col name : ([original col name] , [tbl name] , [col structure])}}
            for tbl in columnTblDic:
                if tbl not in config.QUERY_ALL_COLUMNS_KEY:
                    fullTableName                   = tbl
                    allTableStrucure[tbl.lower()]   = {}
                    if 'schema' in columnTblDic[tbl] and columnTblDic[tbl]['schema'] and len (columnTblDic[tbl]['schema'])>0:
                        fullTableName = columnTblDic[tbl]['schema']+"."+tbl

                    sql = getattr(queries, self.cType + "_columnDefinition")(fullTableName)
                    self.__executeSQL(str (sql), commit=False )

                    for row in self.cursor.fetchall():
                        allTableStrucure[tbl.lower()][row[0].lower()] = ( decodeStrPython2Or3 (row[0], un=True), decodeStrPython2Or3 (tbl, un=True), decodeStrPython2Or3 (row[1].lower().strip().replace(' ', ''), un=True)  )

            # Create source mapping -> tableStructure
            # update mappingDic if there is column mapping

            for i, col in enumerate( alldistinctColumn ):
                targetName = allColumnsTarget[i]
                colType, colTbl, colName = self.__sqlQueryMappingHelp (allTableStrucure, col[1])
                if colName:
                    fullColName = unicode(colName + u"_" + colTbl) if alldistinctColumn.count(colName) > 1 else unicode(colName)
                    if len(col[0])>0:
                        fullColName = col[0]+"."+fullColName
                        colName = col[0]+"."+colName
                    tableStructure.append((unicode(fullColName), colType))
                    # update stt dictionary, if there is a mapping frorm query
                    sttTemp[targetName] = {"s":colName,"t":colType}

            # there is query with * - will add all columns

            if len (alldistinctColumn)<1:
                for tblName in allTableStrucure:
                    if tblName not in [config.QUERY_ALL_COLUMNS_KEY, config.QUERY_SEQ_TAG_VALUE, config.QUERY_SEQ_TAG_FIELD, config.QUERY_TARGET_COLUMNS]:
                        for colTup in allTableStrucure[tblName]:
                            if len (allTableStrucure[tblName][colTup])==3:
                                colName = allTableStrucure[tblName][colTup][0]
                                colType = allTableStrucure[tblName][colTup][2]
                                fullColName = tblName+"."+colName

                                tableStructure.append((unicode(fullColName), colType))
                                sttTemp[colName] = {"s": colName, "t": colType}

            self.cColumns = tableStructure

            for k in sttTemp:
                if stt and k in stt:
                    sttVal = stt[k]
                    if "s" not in sttVal:   sttVal["s"] = sttTemp[k]["s"]
                    if "t" not in sttVal:   sttVal["t"] = sttTemp[k]["t"]

                    sttTemp[k] = sttVal
            if stt:
                for k in stt:
                    if k not in sttTemp: sttTemp[k]=stt[k]
            #Sort stt by query mapping (order by sort if there is a use of more than one source column ....
            if sttTemp and len(sttTemp) > 0 and config.QUERY_SORT_BY_SOURCE:
                listSrc     = []
                listTar     = []
                sttTemp2    = OrderedDict()
                listItems   = sttTemp.items()
                for item in listItems:
                    listTar.append (item[0])
                    if "s" in item[1]:
                        listSrc.append ( item[1]["s"] )
                    else:
                        listSrc.append( None )

                for k in sttTemp:
                    if k not in sttTemp2:
                        sttTemp2[k] = sttTemp[k]
                        if "s" in sttTemp[k]:
                            src = sttTemp[k]["s"]
                            indices = [i for i,x in enumerate(listSrc) if x==src ]
                            if len(indices)>1:
                                for j in indices[1:]:
                                    addKey = listTar[j]
                                    sttTemp2[addKey] = sttTemp[addKey]
                sttTemp = sttTemp2

        sttTemp = sttTemp if sttTemp and len(sttTemp)>0 else None
        if not addSourceColumn and stt and len(stt)>0:
            sttTemp = stt
        return sttTemp

    def __wrapSql (self, col, remove=False, cType=None):
        if cType:
            srcPre, srcPost = config.DATA_TYPE['colFrame'][cType]
        else:
            srcPre, srcPost = config.DATA_TYPE['colFrame'][self.cType]
        coList = col.split(".")
        ret = ""
        for col in coList:
            col = col.replace(srcPre,"").replace(srcPost,"")
            if not remove:
                col= "%s%s%s" %(srcPre,col,srcPost)
            ret+=col+"."
        return ret[:-1]

    # Neeeds to support unicode values...
    def __access(self,tableName):
        ret = []
        for row in self.cursor.columns():
            if len(row)>3:
                curTblName = row[2]
                curTblName = curTblName.encode("utf-8")
                if curTblName == tableName:
                    colName = row.column_name.encode("utf-8")
                    colType = 'varchar(100)'
                    aType = row.type_name.lower()
                    if aType in ('varchar', 'longchar', 'bit','ntext'):
                        if row.column_size > 4098:
                            colType = 'varchar(max)'
                        else:
                            colType = 'varchar(' + str(row.column_size) + ')'
                    elif aType in ('integer', 'counter'):
                        colType = 'int'
                    elif aType in ('double'):
                        colType = 'float'
                    elif aType in ('decimal'):
                        colType = 'decimal(' + str(row.column_size) + "," + str(row.decimal_digits) + ")"
                    ret.append ( (colName , colType) )
        return ret



# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of popEye
#
# popEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# popEye is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cadenceEtl.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import  (absolute_import, division, print_function)
__metaclass__ = type

import collections
import os
import shutil
import time
import csv
import io
import codecs
import pandas as pd
from collections import OrderedDict

from popEtl.config          import config
from popEtl.glob.glob       import p
from popEtl.glob.enums      import eConnValues, eDbType


class cnFile ():
    def __init__ (self, connDic, connName=None, connFolder=None, fileHeader=[]):
        self.cName      = connDic [ eConnValues.connObj ]  if connDic else connName
        self.cType      = connDic [ eConnValues.connType ] if connDic else eDbType.FILE
        self.folderPath = connDic [ eConnValues.connUrl ]  if connDic else connFolder
        self.fileHeader = fileHeader
        self.cursor     = None
        self.conn       = None
        self.cColumns   = None

        if isinstance(self.folderPath, (dict, OrderedDict)):
            connProp = self.folderPath.keys()
        self.fileDelimiter  = self.folderPath['delimiter']  if 'delimiter'   in connProp    else config.FILE_DEFAULT_DELIMITER
        self.fileHeader     = self.folderPath['header']     if 'header' in connProp         else config.FILE_DEFAULT_HEADER
        self.folderPath     = self.folderPath['folder']     if 'folder' in connProp         else config.FILE_DEFAULT_FOLDER
        self.fullPath       = os.path.join(self.folderPath, self.cName)
        self.newLine        = self.folderPath['newLine']    if 'newLine' in connProp        else config.FILE_DEFAULT_NEWLINE
        self.encoding       = self.folderPath['encoding']   if 'encoding' in connProp       else config.FILE_DECODING
        self.errors         = self.folderPath['errors']     if 'errors' in connProp         else config.FILE_LOAD_WITH_CHAR_ERR

        head, tail = os.path.split (self.cName)
        if head and len(head)>1 and tail and len (tail)>1:
            self.fullPath = self.cName
        else:
            self.fullPath = os.path.join(self.folderPath, self.cName)
        p ("file-> INIT: %s, Delimiter %s, Header %s " %(str(self.fullPath) , str(self.fileDelimiter) ,str(self.fileHeader) ), "ii")

    def close (self):
        pass

    def truncate(self, tbl=None):
        if os.path.isfile(self.fullPath):
            os.remove(self.fullPath)
            p("file->truncate: %s is deleted " %(self.fullPath))
        else:
            p("file->truncate: %s is not exists " % (self.fullPath))

    def getColumns (self):
        if self.cColumns and len(self.cColumns)>0:
            return self.cColumns
        else:
            self.structure (stt=None)
        return self.cColumns

    def setColumns(self, colList):
        columnsList = []
        ret = []
        # check if column object is ordered dictionay
        if len (colList) == 1 and isinstance( colList[0] , collections.OrderedDict ):
            columnsList = colList[0].items()
        else:
            if isinstance( colList, list) and len (colList)>0:
                columnsList = colList
            else:
                if isinstance( colList, (dict, collections.OrderedDict) ):
                    columnsList = colList.items()
                else:
                    p ("file->setColumns: List of column is not ordered dictioany or list or regualr dictioanry ....","e")
                    return None

        for col in columnsList:
            if (isinstance (col, (tuple, list))):
                colName = col[0]
            else:
                colName = col
            ret.append(colName)
        self.columns = ret

        p("file-> setColumns: type: %s, file %s will be set with column: %s" % (self.cType, self.cName, str(self.columns)), "ii")
        return self.columns

    def create(self, colList, fullPath=None,  seq=None):
        fullPath = fullPath if fullPath else self.fullPath

        if seq:
            p ("file->create: FILE %s, Sequence is not activated in target file connection, seq: %s  ..." %(str(fullPath) , str (seq) ), "e")
        self.__cloneObject()
        if self.fileHeader:
            p ("file->create: FILE %s, using columns %s as hedaers ..." %(str(fullPath) , str(self.fileHeader) ), "ii")
        else:
            p ("file->create: FILE %s, using columns %s as hedaers ..." %(str(fullPath) , str(colList) ) , "ii")

        # create new File
        self.fileObj = open (fullPath, 'w')

    def structure(self, stt ,addSourceColumn=False,tableName=None, sqlQuery=None):
        stt = collections.OrderedDict() if not stt else stt
        addToStt = False
        if (os.path.isfile(self.fullPath)):
            retWithHeaders  = []
            retNoHeaders    = []
            p ('file->structure: file %s exists, delimiter %s, will extract column structure' %( self.fullPath, str(self.fileDelimiter) ), "ii")
            with io.open(self.fullPath, 'r', encoding=config.FILE_DECODING) as f:
                headers = f.readline().strip(config.FILE_DEFAULT_NEWLINE).split(self.fileDelimiter)

            if len(headers)>0:
                defDataType = config.DATA_TYPE['default'][self.cType]
                defColName =  config.FILE_DEF_COLUMN_PREF
                sttSource = {}
                if len(stt)>0:
                    for t in stt:
                        if "s" in stt[t]:
                            if stt[t]["s"] not in sttSource:
                                sttSource[stt[t]["s"]] = config.DATA_TYPE['default'][self.cType]
                                if "t" in stt[t]:
                                    sttSource[stt[t]["s"]] = stt[t]["t"]
                            else:
                                if "t" in stt[t]:
                                    sttSource[stt[t]["s"]] = stt[t]["t"]
                        if "t" not in stt[t]:
                            stt[t]["t"] = config.DATA_TYPE['default'][self.cType]
                else:
                    addToStt = True

                for i , col in enumerate (headers):
                    cName = col if self.fileHeader else config.FILE_DEF_COLUMN_PREF+str(i)
                    cType = config.DATA_TYPE['default'][self.cType]
                    if col in sttSource:
                        cType = sttSource[col]

                    if addSourceColumn or addToStt:
                        if col not in sttSource:
                            stt[cName] = {"s":cName, "t":cType}

                    retWithHeaders.append( (cName , cType) )

                self.cColumns = retWithHeaders
                if (self.fileHeader):
                    p ('file->structure: file %s contain header will use default %s as data type for each field >>> ' %( self.fullPath, str(defDataType) ), "ii")
                else:
                    p ('file->structure: file %s come without headers, will use prefix name %s and default %s as data type for each field >>> ' %( self.fullPath, str(defColName), str(defDataType) ), "ii")
            else:
                p ('file->structure: file %s is empty, there is no mapping to send >>> ' %( str(self.fullPath) ), "ii")
        else:
            p('file->structure: file %s is not exists >>> ' % (str(self.fullPath)), "ii")

        return stt

    def loadData(self, srcVsTar, results, numOfRows, cntColumn):
        headerList = None
        if self.fileHeader:
            if srcVsTar:
                headerList = [t[1] for t in srcVsTar]
            else:
                headerList = ["col%s" %i for i in range ( cntColumn ) ]

        with codecs.open( filename=self.fullPath, mode='wb', encoding="utf8") as f:
            if headerList:
                f.write (u",".join(headerList))
                f.write("\n")

            for row in results:
                row = [unicode(s)  for s in row]
                f.write(u",".join(row))
                f.write("\n")

        p('file->loadData: Load %s into target: %s >>>>>> ' % (str(numOfRows), self.fullPath), "ii")
        return

    def _updateRow (self, row ):
        def rep (s):
            return s.replace('"','').replace ("\t","")
        ret = row
        ret = [rep(row[c]) if str(c).isdigit() and len(row[c])>0 else None for c in self.header]
        if self.fnDic:
            lenR = len (ret)
            for pos, fnList in self.fnDic.items():
                if not isinstance(pos, tuple):
                    uColumn = ret[pos] if lenR<pos else None
                    for f in fnList:
                        uColumn = f.handler(uColumn)
                    if lenR<=pos:
                        ret.append(uColumn)
                    else:
                        ret[pos] = uColumn
                else:
                    fnPos = fnList[0]
                    fnStr = fnList[1]
                    fnEval = fnList[2]
                    newVal = [str(ret[cr]).decode(config.FILE_DECODING) for cr in pos]
                    newValStr = str(fnStr,'utf-8').format(*newVal)
                    if lenR<=pos:
                        ret.append( eval(newValStr) if fnEval else newValStr )
                    else:
                        ret[fnPos] = eval(newValStr) if fnEval else newValStr
        return ret

    def dfToTable(self, df, ifExists='truncate', seq=None, index=False,chunksize=None):
        p("file->dfToTable: tranfering to file %s from data frame  >>>" % self.fullPath, "ii")

        if seq:
            p ("file->dfToTable: FILE %s, Sequence is not activated in target file connection, seq: %s  ..." %(str(self.fullPath) , str (seq) ), "e")

        if ifExists=='truncate':
            p("file->dfToTable: APPEND DATA  >>>" % self.fullPath, "ii")
            with open(self.cName, 'a') as f:
                df.to_csv(f , header=False, index=index, encoding=config.FILE_ENCODING, quoting=csv.QUOTE_NONNUMERIC,  quotechar = self.fileDelimiter)
        else:
            p("file->dfToTable: TRUNCATE AND LOAD DATA  >>>" % self.fullPath, "ii")
            df.to_csv(self.cName , header=False, index=index, encoding=config.FILE_ENCODING, quoting=csv.QUOTE_NONNUMERIC,  quotechar = self.fileDelimiter)

    def dfFromTable(self,srcColumns=None,tarColumn=None, index_col=None):
        p("file->dfFromTable: loading from file: %s into dataframe >>>" %self.fullPath, "ii")
        if self.fileHeader:
            df = pd.read_csv(filepath_or_buffer=self.fullPath, sep=self.fileDelimiter, encoding=config.FILE_DECODING, keep_default_na=False,
                             index_col=index_col, na_values=config.DATA_TYPE['null'][self.cType])  # , encoding="utf8" names=columns,

        else:
            df = pd.read_csv(filepath_or_buffer=self.fullPath, sep=self.fileDelimiter, encoding=config.FILE_DECODING, header=None, keep_default_na=False,
                             index_col=index_col, na_values=config.DATA_TYPE['null'][self.cType])  # , encoding="utf8" names=columns,

        if srcColumns:
            dfColumn = df.columns
            for i,col in enumerate (tarColumn):
                if srcColumns[i] in dfColumn:
                    df.rename(columns={srcColumns[i]: col}, inplace=True)
                else:
                    df.insert (i , col, '')

        return df.values.tolist()

    def __cloneObject(self, fullPath=None):
        fullPath = fullPath if fullPath else self.fullPath
        fileName = os.path.basename(fullPath)
        fileDir  = os.path.dirname(fullPath)
        fileNameNoExtenseion    = os.path.splitext(fileName)[0]
        fimeNameExtension       = os.path.splitext(fileName)[1]
        ### check if table exists - if exists, create new table
        isFileExists = os.path.isfile(fullPath)
        toUpdateFile = True


        if config.TABLE_HISTORY:
            p ("file-> __cloneObject: FILE History is ON ...", "ii")
            if isFileExists:
                actulSize = os.stat(fullPath).st_size
                if  actulSize<config.FILE_MIN_SIZE:
                    p("file-> __cloneObject: FILE %s exists, file size is %s which is less then %s bytes, will not update ..." % (fullPath, str(actulSize), str(config.FILE_MIN_SIZE)), "ii")
                    toUpdateFile = False
                else:
                    p("file-> __cloneObject: FILE %s exists, file size is %s which is bigger then %s bytes, file history will be kept ..." % (fullPath, str(actulSize), str(config.FILE_MIN_SIZE)), "ii")

            if toUpdateFile:
                oldName = None
                if (os.path.isfile(fullPath)):
                    oldName = fileNameNoExtenseion+"_"+str (time.strftime('%y%m%d'))+fimeNameExtension
                    oldName = os.path.join(fileDir, oldName)
                    if (os.path.isfile(oldName)):
                        num = 1
                        oldName= os.path.splitext(oldName)[0] + "_"+str (num) + os.path.splitext(oldName)[1]
                        oldName = os.path.join(fileDir, oldName)
                        while ( os.path.isfile(oldName) ):
                            num += 1
                            FileNoExt   = os.path.splitext(oldName)[0]
                            FileExt     = os.path.splitext(oldName)[1]
                            oldName=FileNoExt[: FileNoExt.rfind('_') ]+"_"+str (num) + FileExt
                            oldName = os.path.join(fileDir, oldName)
                if oldName:
                    p ("file-> __cloneObject: File History is ON, file %s exists ... will copy this file to %s " %(str (self.cName) , str(oldName) ), "ii")
                    shutil.copy(fullPath, oldName)
        else:
            if ( os.path.isfile(fullPath) ):
                os.remove(fullPath)
                p ("file-> __cloneObject: File History is OFF, and file %s exists, DELETE FILE >>>> " %(str (self.cName)  ), "ii")
            else:
                p ("file-> __cloneObject: File History is OFF, and file %s is not exists, continue >>>> " %(str (self.cName)  ), "ii")



class eDbType (object):
    SQL     = "sql"
    ORACLE  = "oracle"
    VERTIVA = "vertica"
    ACCESS  = "access"
    MYSQL   = "mysql"
    FILE    = "file"

def isDbType (prop):
    dicClass = eDbType.__dict__
    for p in dicClass:
        if isinstance(dicClass[p], str) and dicClass[p].lower() == prop.lower():
            return prop.lower()
    return None


class eConnValues (object):
    connName        = "name"
    connType        = "type"
    connUrl         = "url"
    connUrlExParams = "urlExParams"
    connObj         = "object"
    connIsSql       = "isSql"
    connFilter      = "filter"
    connIsTar       = "isTarget"
    connIsMerge     = "isMerge"
    connIsSrc       = "isSource"
    partitionCol    = "column"
    partitionAgg    = "agg"
    partitionStart  = "start"
    fileToLoad      = "file"

class ePopEtlProp (object):
    src = "source"
    tar = "target"
    qry = "query"
    mrg = "merge"
    add = "addSrcColumns"
    seq = "seq"
    stt = "stt"
    sttA= "sttappend"
    map = "mapping"
    col = "column"
    par = "partition"
    inc = "incremental"
    exe = "execsql"

    dicOfProp = {
        src : ["source","src"],
        tar : ["target","tar"],
        qry : ["query"],
        mrg : ["merge"],
        seq : ["seq"],
        stt : ['stt', 'sttappend'],
        map : ['mapping', 'map'],
        col : ['columns', 'column', 'col'],
        par : ['partition'],
        inc : ['inc', 'incremental'],
        exe : ['esql', 'execsql']
    }


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of popEye
#
# popEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# popEye is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cadenceEtl.  If not, see <http://www.gnu.org/licenses/>.

import sqlparse
from sqlparse.sql       import IdentifierList, Identifier
from sqlparse.tokens    import Keyword, DML

from popEtl.config      import config
from popEtl.glob.glob   import p, replaceStr

def is_subselect(parsed):
    if not parsed.is_group:
        return False
    for item in parsed.tokens:
        if item.ttype is DML and item.value.upper() == 'SELECT':
            return True
    return False

def extract_from_part(parsed):
    from_seen = False
    for item in parsed.tokens:
        if item.is_group:
            for x in extract_from_part(item):
                yield x
        if from_seen:
            if is_subselect(item):
                for x in extract_from_part(item):
                    yield x
            elif item.ttype is Keyword and item.value.upper() in ['ORDER', 'GROUP', 'BY', 'HAVING']:
                from_seen = False
                StopIteration
            else:
                yield item
        if item.ttype is Keyword and item.value.upper() == 'FROM':
            from_seen = True

def extract_select_part (parsed):
    allColumnSign       = config.QUERY_ALL_COLUMNS_KEY
    nonColumnSign       = config.QUERY_COLUMNS_KEY
    targetColumnNames   = config.QUERY_TARGET_COLUMNS
    columnList          = []
    columnDic           = {allColumnSign:[],targetColumnNames:[]}
    col                 = None
    addToken            = False

    for item in parsed.tokens:
        if item.ttype is DML and item.value.upper() == 'SELECT':
            addToken = True

        elif item.ttype is Keyword and item.value.upper() == 'FROM':
            addToken = False
            break
        else:
            if addToken:
                dicKey  = None
                dicValue= None

                if isinstance(item, IdentifierList):
                    for identifier in item.get_identifiers():
                        identifier = str(identifier)
                        srcName = identifier

                        if identifier.lower().find(" as") > 0:
                            srcName = identifier[:identifier.lower().find(" as")].strip()
                            tarName = identifier[identifier.lower().find(" as")+3:].strip()
                        else:
                            tarName = srcName.split(".")
                            tarName = tarName[1] if len(tarName)>1 else tarName[0]

                        columnList.append( (srcName.split(".") , tarName) )
                elif isinstance(item, Identifier):
                    item = str(item)
                    srcName = item
                    if item.lower().find(" as") > 0:
                        srcName = item[:item.lower().find(" as")].strip()
                        tarName = item[item.lower().find(" as")+3:].strip()
                    else:
                        tarName = srcName
                    columnList.append( (srcName.split(".") , tarName) )

    for tupCol in columnList:

        col     = tupCol[0]
        tarName = tupCol[1]
        if col and len (col) == 1:
            dicKey = nonColumnSign
            dicValue = col[0]
        elif col and len (col) >= 2:
            dicKey = col[0]
            dicValue = ".".join(col)
            dicValue = dicValue.replace("\n", " ")
            if dicKey not in columnDic:
                columnDic[dicKey] = []
        else:
            p("dbQueryParser->extract_select_part: ERROR Loading column identifier, will ignore column %s" % str(col), "i")
            continue

        columnDic[allColumnSign].append( dicValue )
        columnDic[targetColumnNames].append (tarName)
        if dicKey and dicKey not in columnDic:
                columnDic[dicKey] = []
        columnDic[dicKey].append (dicValue)

    return columnDic


def extract_table_identifiers(token_stream):
    for item in token_stream:
        if isinstance(item, IdentifierList):
            for identifier in item.get_identifiers():
                value = ( identifier.get_alias() , identifier._get_first_name(), identifier.get_real_name())
                value = [x.replace('"', '').replace("'", "") if x else None for x in  value]
                value = tuple( value )
                yield value

        elif isinstance(item, Identifier):
            value = (item.get_alias(), item._get_first_name(), item.get_real_name())
            value = [x.replace('"', '').replace("'", "") if x else None for x in value]
            value = tuple(value)
            yield value

def extract_tables(sql):
    # let's handle multiple statements in one sql string
    extracted_tables = []
    extracted_last_tables_Tuple = None
    extracted_last_columns_Dic = None
    # replacements to SQL queries
    sql = replaceStr (sString=sql,findStr="ISNULL (", repStr="ISNULL(", ignoreCase=True,addQuotes=None)
    sql = replaceStr(sString=sql, findStr="CONVERT (",repStr="CONVERT(", ignoreCase=True, addQuotes=None)
    sql = sql.replace("\t"," ")
    statements = list(sqlparse.parse(sql))

    for statement in statements:
        if statement.get_type() != 'UNKNOWN':
            stream = extract_from_part(statement)
            # will get only last table column definistion !
            extracted_last_columns_Dic  = extract_select_part(statement)
            extracted_last_tables_Tuple = list (extract_table_identifiers(stream))
            #extracted_tables.append(list(extract_table_identifiers(stream)))
    #return list(itertools.chain(*extracted_tables))
    return (extracted_last_tables_Tuple , extracted_last_columns_Dic)

def extract_tableAndColumns (sql):
    ret = {}
    tblTupe , columns = extract_tables(sql)

    if config.QUERY_ALL_COLUMNS_KEY in columns:
        ret[config.QUERY_ALL_COLUMNS_KEY] = columns[config.QUERY_ALL_COLUMNS_KEY]
    if config.QUERY_TARGET_COLUMNS in columns:
        ret[config.QUERY_TARGET_COLUMNS] = columns[config.QUERY_TARGET_COLUMNS]
    # merge table to columns

    for tbl in tblTupe:
        alias       = tbl[0]
        schamenName = tbl[1]
        tableName   = tbl[2]

        if tableName not in ret:
            ret[tableName] = {'alias':alias, 'schema':None if schamenName==tableName else schamenName}

        if columns and len (columns)>0:
            ret[tableName]['column'] = []

        for col in columns:
            if str(col) in [tableName,alias]:
                ret[tableName]['column'].extend  (columns[col])
            if str(col) == config.QUERY_COLUMNS_KEY:
                ret[tableName]['column'].extend (columns[col])

    return ret


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of popEye
#
# popEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# popEye is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cadenceEtl.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import json
import os
import re
import io
from collections            import OrderedDict

from popEtl.config          import config
from popEtl.glob.glob       import p, setDicConnValue, getDicKey, filterFiles
from popEtl.glob.enums      import eConnValues, ePopEtlProp
from popEtl.glob.globalDBFunctions import checkSequence, logsToDb
from popEtl.connections.connector  import connector

# mapping - change source data type to destination data type
def sourceToTargetDataTypes (srcType, trgType, srcColumns):
    if "file" in srcType.lower():
        srcType = trgType
    newColumns  = []
    mappDic     = {}
    ret         = None

    if "access" in srcType: srcType="sql"

    # source is same as target, data type ios the same
    if srcType == trgType:
        return srcColumns

    # loop on config.DATA_TYPE and create dictionary with source to target
    for dType in config.DATA_TYPE:
        srcTypes = None
        trgTypes = None
        if srcType in  config.DATA_TYPE[dType]: srcTypes = config.DATA_TYPE[dType][srcType]

        if trgType in config.DATA_TYPE[dType]: trgTypes = config.DATA_TYPE[dType][trgType]

        if srcTypes and trgTypes:
            if isinstance(srcTypes, tuple):
                for s in srcTypes:
                    if s not in mappDic:
                        mappDic[s] = trgTypes[0] if isinstance(trgTypes, tuple) else trgTypes
            else:
                if srcTypes not in mappDic:
                    mappDic[srcTypes] = trgTypes[0] if isinstance(trgTypes, tuple) else trgTypes

    if isinstance( srcColumns, (dict, OrderedDict) ):
        listTarToSrc = [(x,srcColumns[x]["t"]) for x in srcColumns if "t" in srcColumns[x]]
    else:
        listTarToSrc = srcColumns

    # update pre/post column
    # update column with targer data type
    for col in listTarToSrc:
        columnName      = col[0]
        columnType      = col[1].replace(' ','')
        replaceString   = ""
        postType        = None
        # metch string(...) pattern


        fmatch = re.search (r'(.*)(\(.+\))',columnType, re.M|re.I)
        newField = config.DATA_TYPE['default'][trgType]
        if fmatch:
            replaceString   = fmatch.group(1)   # --> varchar, int , ...
            postType        = fmatch.group(2)   # --> (X), (X,X) , ....
        else:
            replaceString = columnType

        if replaceString in mappDic:
            newField = mappDic[replaceString]
            if postType:
                # for my sql only !! - int(11) will be considered as int ...
                newField +=postType if 'int' not in replaceString else ""
        newColumns.append ( (columnName , newField) )

    if isinstance( srcColumns, (dict, OrderedDict) ):
        for tup in newColumns:
            existsVal = srcColumns[tup[0]]
            existsVal["t"] = tup[1]
            srcColumns[tup[0]] = existsVal
        ret = srcColumns
    else:
        ret = newColumns

    p ("mapper->sourceToTargetDataTypes: source data types converted to destination. new dataTpe: %s >>>>" %(str(newColumns)) ,"ii")
    return ret

# dictionary with keys : 'source','target','columns','mapping'
def mapper (dicProp):
    prop        = dicProp.keys()
    toCreate    = []

    #update parmaters
    isMerge         = dicProp[ePopEtlProp.mrg] if ePopEtlProp.mrg in prop else None
    stt             = dicProp[ePopEtlProp.stt] if ePopEtlProp.stt in prop else None
    seq             = checkSequence (dicProp[ePopEtlProp.seq]) if ePopEtlProp.seq in prop else None
    isTarget        = dicProp[ePopEtlProp.tar] if ePopEtlProp.tar in prop else None
    isSource        = dicProp[ePopEtlProp.src] if ePopEtlProp.src in prop else None
    addSourceColumn = dicProp[ePopEtlProp.add] if ePopEtlProp.add in prop else False

    # There is Target mapping
    if not isTarget:
        p("mapper->mapper: Target is not exists, What the hell quittttitnnggggg.... >>>>>>>>>>", "i")
        return

    # There is direct source
    if isSource:
        stt         = isSource.structure(stt=stt,addSourceColumn=addSourceColumn)
        srcColumns  = isSource.getColumns()

        # convert source data type to target data types
        srcColumns      = sourceToTargetDataTypes(isSource.cType, isTarget.cType, srcColumns)
        srcColumnNames  = [c[0].strip() for c in srcColumns]

        updColumns      = []
        if srcColumns and len (srcColumns)>0:
            # there is target, source and mapping Will map field based on source and create table
            if stt and len(stt)>0:
                srcC = [stt[k]["s"] for k in stt if "s" in stt[k]]

                # check if all mapping exist in source as well
                for col in srcC:
                    if col not in srcColumnNames:
                        colTodel = [k for k in stt if "s" in stt[k] and stt[k]["s"]==col]
                        p ('mapper->mapper: There is Targets %s which is mapped to %s, but source not exists, ignoring targets column %s ....' %(str(colTodel), col,str(colTodel) ) ,"ii")
                        p('mapper->mapper: SOURCE COLUMNS : '  , "ii")
                        srcNameStr = ""
                        for s in srcColumnNames:
                            srcNameStr+=s+" ,"
                        p(srcNameStr, "ii")
                        for d in colTodel:
                            del stt[d]

                # Update stt with new types
                stt = sourceToTargetDataTypes(isSource.cType, isTarget.cType, stt)

            isTarget.create(stt=stt, seq=seq)
            if isMerge: isMerge.create(stt=stt, seq=seq)
        else:
            p ("mapper->mapper: Source %s is not exists, will not create target table >>>>>>>>>>" %str(isSource.cName),"e")
    # there is only target
    elif stt:
        for t in stt:
            if "t" not in stt[t]:
                p('mapper->mapper: There is Target mapping without Source, there is NO TYPE for column %s, ignoring column, values is %s ....' % (str(t), str(stt[t])), "ii")
                del stt[t]
            if "s" not in stt[t]:
                stt[t]["s"]=t
        isTarget.create(stt=stt, seq=seq)
    elif isMerge:
        stt = isTarget.structure(stt=stt, addSourceColumn=addSourceColumn)
        isMerge.create (stt=stt,seq=seq)

def _extractNodes (jText,jFileName,sourceList=None, destList=None):
    sttDic = OrderedDict()
    for jMap in jText:
        toLoad = True
        dicProp = {}
        srcObj  = None
        tarObj  = None
        sttDic  = OrderedDict()
        dicProp[ ePopEtlProp.add ] = False

        if isinstance(jMap, (list,tuple)):
            _extractNodes(jText=jMap, jFileName=jFileName, sourceList=sourceList, destList=destList)
            continue

        keys = [x.lower() for x in jMap.keys()]
        # update all variables
        sourceConn      = getDicKey(ePopEtlProp.src,keys)
        queryConn       = getDicKey(ePopEtlProp.qry,keys)
        targetConn      = getDicKey(ePopEtlProp.tar,keys)
        mergeConn       = getDicKey(ePopEtlProp.mrg,keys)
        seqFiles        = getDicKey(ePopEtlProp.seq,keys)
        stt             = getDicKey(ePopEtlProp.stt,keys)
        targetMapping   = getDicKey(ePopEtlProp.map,keys)
        targetColumn    = getDicKey(ePopEtlProp.col,keys)

        if sourceConn:
            connDic = setDicConnValue (connJsonVal=jMap[sourceConn],extraConnVal=jFileName, isSource=True )
            srcObj  = connDic [ eConnValues.connObj ] if connDic else None
            dicProp[ePopEtlProp.src] = connector(connDic = connDic)

        if queryConn:
            connDic = setDicConnValue(connJsonVal=jMap[queryConn], extraConnVal=jFileName, isSource=True, isSql=True)
            dicProp[ePopEtlProp.src] = connector(connDic = connDic)
            if sourceConn:
                p("mappr->_extractNodes: There is query and source, will be QUERY sed as SOURCE object >>>>>> ", "ii")

        # target -> Connection object
        if targetConn:
            connDic = setDicConnValue(connJsonVal=jMap[targetConn], connObj=tarObj, extraConnVal=jFileName,  isTarget=True)
            tarObj = connDic[eConnValues.connObj] if connDic else None
            dicProp[ePopEtlProp.tar] = connector(connDic = connDic)

        # create table with sequence as fisrt column -> apply on target only ! (if there is merge - will use merge option, without adding identity
        if seqFiles:
            # check sequence dictionay
            dicProp[ePopEtlProp.seq] = checkSequence(jMap[seqFiles])
            # add property - if merge is appear
            if mergeConn and ePopEtlProp.seq in dicProp and len(dicProp[ePopEtlProp.seq]) > 0:
                dicProp[ePopEtlProp.seq][ePopEtlProp.mrg] = True

        # Check stt (source to target) --> if needs to add columns or new mapping
        if stt:
            sttDic = jMap[stt]
            dicProp[ePopEtlProp.stt] = sttDic
            if ePopEtlProp.sttA in stt:  dicProp[ePopEtlProp.add] = True

        # Mapping source to target fields
        if targetMapping:
            dicProp[ePopEtlProp.add] = False
            if targetConn and (sourceConn or queryConn) :
                mapping = jMap[targetMapping]
                sttDicTemp = sttDic
                sttDic = OrderedDict()
                for m in mapping:
                    if m in sttDicTemp:
                        sttDic[m] = sttDicTemp[m]
                        if "s" not in sttDic[m]:
                            sttDic[m]["s"] = mapping[m]
                    else:
                        sttDic[m] = {"s": mapping[m]}
                for t in sttDicTemp:
                    if t not in sttDic: sttDic[t] = sttDicTemp[t]
                dicProp[ePopEtlProp.stt] = sttDic
            else:
                p("mappr->loadJson: Mapping exists, but there is no target or source connection ... nothing to do ...","e")

        # Create target column by column definitions
        if targetColumn:
            if targetConn:
                columns = jMap[targetColumn]
                sttDicTemp = sttDic
                sttDic = OrderedDict()
                for c in columns:
                    if c in sttDicTemp:
                        sttDic[c] = sttDicTemp[c]
                        if "t" not in sttDic[c]:
                            sttDic[c]["t"] = columns[c]
                    else:
                        sttDic[c] = {"t": columns[c]}
                for t in sttDicTemp:
                    if t not in sttDic: sttDic[t] = sttDicTemp[t]
                dicProp[ePopEtlProp.stt] = sttDic
                dicProp[ePopEtlProp.tar].setColumns(sttDic)
            else:
                p("mappr->loadJson: Target columns exists, but there is no target connection ... nothing to do ...","e")

        # checl list of source or target to load (if list exists)
        if sourceList:
            sourceList = [x.lower() for x in sourceList]
            toLoad = True if ePopEtlProp.src in dicProp and dicProp[ePopEtlProp.src].cName in sourceList else False
        if destList:
            destList = [x.lower() for x in destList]
            toLoad = True if ePopEtlProp.tar in dicProp and dicProp[ePopEtlProp.tar].cName in destList else False

        if toLoad:
            mapper(dicProp=dicProp)
        else:
            p("mappr->loadJson: Src %s, dst %s , mapping %s not matched >>>> nothing to map " % (
            str(sourceList), str(destList), str(dicProp)), "i")

# Main function : loading all json file and parse them by definition
def model (dicObj=None, sourceList=None, destList=None):
    p('mapper->model: START MAPPING >>>>>', "i")

    if dicObj:
        dicObj = list (dicObj) if isinstance(dicObj, (dict,OrderedDict)) else dicObj
        p('mapper->model: loading from Dictionary  >>>>>' , "ii")
        _extractNodes(jText=dicObj, jFileName='', sourceList=sourceList, destList=destList)
    else:
        jsonFiles = filterFiles (modelToExec="mapper->model", dirData=None, includeFiles=None, notIncludeFiles=None )

        for index, js in enumerate(jsonFiles):
            with io.open(os.path.join(config.DIR_DATA, js), encoding='utf-8') as jsonFile:
                p('mapper->model: start modeling from file  %s, folder: %s >>>>>' % (js, str(config.DIR_DATA)), "ii")
                jText = json.load(jsonFile , object_pairs_hook=OrderedDict)
                _extractNodes(jText=jText, jFileName=js, sourceList=sourceList, destList=destList)

    if config.LOGS_IN_DB: logsToDb()
    p ('mapper->model: FINISH MAPPING >>>>>', "i")
    p ('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>', "ii")


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of popEye
#
# popEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# popEye is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cadenceEtl.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import  (absolute_import, division, print_function)
__metaclass__ = type

import os
import sqlite3

from popEtl.glob.glob import p

class sqlLite ():
    def __init__(self, connFile):
        self.db     = None
        self.cur    = None

        self.seqTable = 'seq'
        sqlCreateDB =   [   "create table "+ self.seqTable + " (tbl varchar(100), col varchar(100), sId bigint);",
                            "CREATE UNIQUE INDEX idx_tbl ON " + self.seqTable + " (tbl);"]

        if os.path.isfile(connFile):
            self.db = sqlite3.connect(connFile)
            self.cur = self.db.cursor()
            p("dbSqlLite->init: sqlLite DB is exists, file : %s" % str(connFile), "ii")
        else:
            self.db = sqlite3.connect(connFile)
            self.cur = self.db.cursor()
            self.__execStatement(sqlCreateDB)
            p("dbSqlLite->init: sqlLite DB NOT exists.. creating new one at file : %s" % str(connFile), "i")

    # Get sequence from seq tables
    def execSqlGetValue (self, tblName, colName):
        sql = "select sId from %s where tbl like '%s' and col like '%s';" %(str(self.seqTable), str(tblName), str(colName) )
        row = self.__execSelect(sql, onlyOne=True)
        if row and len(row) > 0:
            newSeq = str(row[0])
            p("dbSqlLite->execSqlGetValue: sqlLite sequence for table: %s, column %s exists, seq: %s... " % (str(tblName), str(colName), newSeq ), "ii")
            return newSeq

        p("dbSqlLite->execSqlGetValue: sqlLite sequence for table: %s, column %s NOT EXISTS... " % (str(tblName), str(colName)), "i")
        return None

    def execReplaceSql (self, tblName, colName, sID):
        sql = "REPLACE INTO %s (tbl, col, sId) VALUES ('%s', '%s', %s);" % (str(self.seqTable), str(tblName), str(colName), str(sID))
        p("dbSqlLite->execReplaceSql: insert or update table: %s, column: %s with new sequence: %s " %(str(tblName), str(colName), str(sID)) ,"ii")
        self.__execStatement(sql)

    def close (self):
        self.db.close()

    def __execStatement(self, sqlList):
        if isinstance(sqlList, (list, tuple)):
            for s in sqlList:
                self.cur.execute(s)
        else:
            self.cur.execute(sqlList)

        self.db.commit()

    def __execSelect(self, sql, onlyOne=False):
        if 'select' in sql.lower():
            self.cur.execute(sql)
            p('dbSqlLite->:__execSelect: SELECT statement ... sql: %s' % (sql), "ii")
        else:
            p('dbSqlLite->:__execSelect: Not a select statement ... sql: %s' %(sql),"ii")
            return
        if onlyOne:
            return self.cur.fetchone()
        return self.cur.fetchall()


# -*- coding: utf-8 -*-
# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of popEye
#
# popEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# popEye is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cadenceEtl.  If not, see <http://www.gnu.org/licenses/>.


######### SQL : RENAME TABLE
def sql_renameTable (tblName, tblNewName):
    return "EXEC sp_rename '"+tblName+"', '"+tblNewName+"'"

######### SQL : DROP TABLE
def sql_dropTable (tblName):
    return "drop table "+tblName

def oracle_dropTable (tblName):
    return "Drop table "+ tblName + ";"

######### SQL : TRUNCATE TABLE
def sql_truncate (tblName):
    return "truncate table "+tblName

def oracle_truncate (tblName):
    return "truncate table "+ tblName + ";"

######### SQL : COLUMN DEFINISTION : column name, column type
def mysql_columnDefinition (tbl):
    schema = None
    endSchema = tbl.find(".",0)
    if endSchema>0:
        schema = tbl[0:endSchema]
        tbl = tbl[endSchema+1:]
    sql = """
    SELECT distinct column_name, column_type  FROM information_schema.columns
    WHERE table_name='"""+tbl+"' "
    sql += "and TABLE_SCHEMA='"+schema+"';" if schema else ";"

    return sql

def vertica_columnDefinition (tbl):
    schema = None
    endSchema = tbl.find(".",0)
    if endSchema>0:
        schema = tbl[0:endSchema]
        tbl = tbl[endSchema+1:]
    # SELECT column_name, data_type, is_nullable FROM columns  WHERE table_schema = 'data' And table_name='historical_status_trials';
    sql = """
    SELECT distinct column_name, data_type  FROM columns
    WHERE table_name='"""+tbl+"' "
    sql += "and table_schema='"+schema+"';" if schema else ";"

    return sql

def sql_columnDefinition (tbl):
    schema = None
    endSchema = tbl.find(".",0)
    if endSchema>0:
        schema = tbl[0:endSchema]
        tbl = tbl[endSchema+1:]

    sql = """
    SELECT c.name,
    UPPER(tp.name) +
    CASE	WHEN tp.name IN ('varchar', 'char', 'varbinary', 'binary', 'text')
                THEN '(' + CASE WHEN c.max_length = -1 THEN 'MAX' ELSE CAST(c.max_length AS VARCHAR(5)) END + ')'
            WHEN tp.name IN ('nvarchar', 'nchar')
                THEN '(' + CASE WHEN c.max_length = -1 THEN 'MAX' ELSE CAST(c.max_length / 2 AS VARCHAR(5)) END + ')'
            WHEN tp.name IN ('ntext')
                THEN ''
            WHEN tp.name IN ('datetime2', 'time2', 'datetimeoffset')
                THEN '(' + CAST(c.scale AS VARCHAR(5)) + ')'
            WHEN tp.name IN ('decimal','numeric')
                THEN '(' + CAST(c.[precision] AS VARCHAR(5)) + ',' + CAST(c.scale AS VARCHAR(5)) + ')'
        ELSE ''
        END as colType

    FROM sys.columns c WITH (NOWAIT)
    JOIN sys.types tp WITH (NOWAIT) ON c.user_type_id = tp.user_type_id
    WHERE c.[object_id] =
        (Select top 1 object_id as obID from
        (SELECT SCHEMA_NAME(schema_id) schemaDesc , name , object_id FROM sys.tables
         Union
        Select SCHEMA_NAME(schema_id) schemaDesc , name , object_id FROM sys.views) tt
        Where """
    sql += "schemaDesc='"+schema.replace("[","").replace("]","")+"' and " if schema else ""
    sql+= "name='"+tbl.replace("[","").replace("]","")+"') ORDER BY c.column_id"

    return sql

def oracle_columnDefinition ( tbl ):
    schema = None
    endSchema = tbl.find(".", 0)

    if endSchema > 0:
        schema = tbl[0:endSchema]
        tbl = tbl[endSchema + 1:]

    sql = "select column_name, data_type || "
    sql += "case when data_type = 'NUMBER' and data_precision is not null then '(' | | data_precision | | ',' | | data_scale | | ')' "
    sql += "when data_type = 'NUMBER' then '(18,0)' "
    sql += "when data_type like '%CHAR%' then '(' | | data_length | | ')' "
    sql += "else '' end type "
    sql += "from all_tab_columns where table_name = '" + tbl + "' "
    sql += " and owner='" + schema + "'" if schema else ""
    sql += " ORDER BY COLUMN_ID"
    return str(sql)

######### SQL : COLUMN NAMES : column name
def sql_columnsNames (tblName, tblSchema=None):
    if not tblSchema:
        endSchema = tblName.find(".", 0)
        if endSchema > 0:
            tblSchema = tblName[0:endSchema]
            tblName = tblName[endSchema + 1:]

    if tblSchema:
        sql = "select column_name from information_schema.columns where table_name='" + tblName + "' And table_schema='" + tblSchema + "' order by ordinal_position"
    else:
        sql = "select column_name from information_schema.columns where table_name='" + tblName + "' order by ordinal_position"
    return sql

def oracle_columnsNames (tblName, tblSchema=None):
    if not tblSchema:
        endSchema = tblName.find(".", 0)
        if endSchema > 0:
            tblSchema = tblName[0:endSchema]
            tblName = tblName[endSchema + 1:]
    if tblSchema:
        sql = "select column_name from information_schema.columns where table_name='" + tblName + "' And table_schema='" + tblSchema + "' order by ordinal_position"
    else:
        sql = "select column_name from information_schema.columns where table_name='" + tblName + "' order by ordinal_position"
    return sql

######### SQL : DATABASE STRUCURE
def sql_objectStrucute (filterDic):
    typesObj    = "'BASE TABLE','VIEW'"
    likeStr     = None
    sql         = ""
    if filterDic:
        if 'type' in filterDic:
            typesObj = ",".join([ "'"+x.replace ('"','').replace("'","")+"'" for x in filterDic['type']])
        if 'like' in filterDic:
            likeStr = " TABLE_NAME like ('%"+filterDic['like'].replace("'","").replace('"','')+"%') "
    # select type_desc,type, name from sys.objects WHERE type in ( %s ) AND %s order by name
    # SELECT TABLE_SCHEMA+'.'+TABLE_NAME FROM INFORMATION_SCHEMA.TABLES Where TABLE_TYPE in ('BASE TABLE','VIEW')
    if  likeStr:
        sql = "SELECT TABLE_SCHEMA+'.'+TABLE_NAME, TABLE_TYPE FROM INFORMATION_SCHEMA.TABLES Where TABLE_TYPE in ( %s ) AND %s " %( typesObj , likeStr)
    else:
        sql = "SELECT TABLE_SCHEMA+'.'+TABLE_NAME, TABLE_TYPE FROM INFORMATION_SCHEMA.TABLES Where TABLE_TYPE in ( %s ) " % (typesObj)
    sql+=" ORDER BY TABLE_SCHEMA+'.'+TABLE_NAME"
    return sql

######### SQL : MINIMUN VALUE : tblName, tblSchema, resolution, periods, col=None, startDate=None
def sql_minValue (tblName, tblSchema, resolution,periods,col=None, startDate=None):
    sql = ""
    dDate = " getdate() "
    if startDate:
        dDate = (" Convert (smalldatetime '%s') " %str(startDate))
    if col:
        sql = "Select CONVERT (DATE, MIN (%s)) FROM " % str (col)
        if tblSchema:
            sql += tblSchema + "." + tblName
        else:
            sql += tblName
    else:
        sql = "Select convert (date, dataadd (%s, %s, %s))" %(str(resolution),str(periods),dDate)
    return sql

def mysql_minValue (tblName, tblSchema, resolution,periods,col=None, startDate=None):
    sql = ""
    if 'd' == resolution: resolution = "DAY"
    if 'm' == resolution: resolution = "MONTH"
    if 'y' == resolution: resolution = "YEAR"

    dDate = " CURDATE() "
    if startDate:
        dDate = (" '%s' " %str(startDate))
    if col:
        sql = "Select DATE ( MIN (%s)) FROM " % str (col)
        if tblSchema:
            sql += tblSchema + "." + tblName
        else:
            sql += tblName
    else:
        sql = "Select DATE ( DATE_ADD(%s, INTERVAL %s %s))" %(dDate,str(periods),str(resolution))
    return sql

def oracle_minValue (tblName, tblSchema, resolution,periods,col=None, startDate=None ):
    sql = ""
    dDate = " getdate() "
    if startDate:
        dDate = (" TO_DATE ('%s') " % str(startDate))
    if col:
        sql = "Select trunc (MIN (%s) ) FROM " % str (col)
        if tblSchema:
            sql += "`"+tblSchema + "`" + "." + "`" + tblName + "`"
        else:
            sql += "`" + tblName + "`"
    if 'd' in resolution:
        sql = 'Select trunc (%s+%s) from dual;' %(dDate, str(periods))
    if 'm' in resolution:
        sql =  'Select trunc (add_months(%s, %s)) from dual;' %(dDate, str(periods))
    if 'y' in resolution:
        sql =  'Select trunc (add_months(%s, %s)) from dual;' %(dDate, str(periods*12))
    return sql

######### SQL : MERGE : dstTable, srcTable, mergeKeys, colList , colFullList
def sql_merge (dstTable, srcTable, mergeKeys, colList , colFullList):
        sql = "MERGE INTO " + dstTable + " as t USING " + srcTable + " as s ON ("
        colOnMerge = " AND ".join(["ISNULL (t." + c + ",'')= ISNULL (s." + c +",'')" for c in mergeKeys])
        sql += colOnMerge + ") \n WHEN MATCHED THEN UPDATE SET \n"
        for c in colList:
            # Merge only is source is not null
            sql += "t." + c + "=" + "case when s." + c + " is null or len(s." + c + ")<1 then t." + c + " else s." + c + " End,\n"
        sql = sql[:-2] + "\n"
        sql += " WHEN NOT MATCHED THEN \n"
        sql += " INSERT (" + ",".join([c for c in colFullList]) + ") \n"
        sql += " VALUES  (" + ",".join(["s." + c for c in colFullList]) + "); "

        return sql

######### SQL : SEQUENCE : column, type, start, leg
def sql_seq (seqDic):
    sql = ""
    if 'column' in seqDic and 'type' in seqDic and 'start' in seqDic and 'inc' in seqDic:
        if 'merge' in seqDic:
            # "["+seqDic['column']+"]"+"\t"+
            sql = "["+seqDic['type']+"],\n"
        else:
            sql = "["+seqDic['type']+"]"+"\t"+" IDENTITY("+str(seqDic['start'])+","+str(seqDic['inc'])+") NOT NULL,\n"
    return sql


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of popEye
#
# popEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# popEye is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cadenceEtl.  If not, see <http://www.gnu.org/licenses/>.

import re
import sys
import os
import io
import datetime
import logging
from collections import OrderedDict

from popEtl.glob.enums import eConnValues, eDbType, ePopEtlProp, isDbType
from popEtl.glob.sqlPythonQueries import queryParsetIntoList
from  popEtl.config import config

def getLogger (
    LOG_FORMAT     = '%(asctime)s %(levelname)s %(message)s',
    LOG_DIR        = None,
    LOG_FILE       = 'file',
    LOGS_DEBUG      = logging.DEBUG
    ):

    #logging.basicConfig(format='%(asctime)s - %(message)s', datefmt='%d-%b-%y %H:%M:%S')
    logFormatter= logging.Formatter(LOG_FORMAT)
    logg  = logging.getLogger()

    if LOG_DIR:
        fileHandler = logging.FileHandler("{0}/{1}.log".format(LOG_DIR, LOG_FILE))
        fileHandler.setFormatter(logFormatter)
        logg.addHandler(fileHandler)

    consoleHandler = logging.StreamHandler(sys.stdout)
    consoleHandler.setFormatter(logFormatter)
    logg.addHandler(consoleHandler)

    logg.setLevel( LOGS_DEBUG )

    return logg

logg = getLogger (LOGS_DEBUG=config.LOGS_DEBUG)
def p(msg, ind='I'):
    ind = ind.upper()
    indPrint = {'E': 'ERROR>> ',
                'I': 'INFO >> ',
                'II': 'DEBUG>> ',
                'III': 'Progress>> '}
    allowToPrint    = ['E', 'I','II', 'III']  #  'II', 'III'
    #allowToPrint = ['E','I']
    allowToSaveInDB_E = ['E']
    allowToSaveInDB_I = ['I']
    allowToSaveInDB = allowToSaveInDB_E + allowToSaveInDB_I

    if ind in allowToPrint or (config.LOGS_IN_DB and ind in allowToSaveInDB):
        localTime = datetime.datetime.today()
        if config.LOGS_IN_DB and ind in allowToSaveInDB_E:
            timeStr = localTime.strftime("%m/%d/%Y %H:%M:%S")
            config.LOGS_ARR_E.append((timeStr,config.LOGS_DB_TIME_STEMP, ind, str(msg)))
        elif config.LOGS_IN_DB and ind in allowToSaveInDB_I:
            timeStr = localTime.strftime("%m/%d/%Y %H:%M:%S")
            config.LOGS_ARR_I.append((timeStr,config.LOGS_DB_TIME_STEMP, ind, str(msg)))
        if config.LOGS_PRINT and ind in allowToPrint:
            timeStr = localTime.strftime("%d/%m/%Y %H:%M:%S")
            if 'III' in ind:
                logg.debug("\r %s %s" %(indPrint[ind], msg))
            elif 'II' in ind:
                logg.debug("%s %s" %(indPrint[ind], msg))
            elif 'I' in ind:
                logg.info("%s %s" %(indPrint[ind], msg))
            else:
                logg.error(str(indPrint[ind]) + str(msg))

def setQueryWithParams(query):
    qRet = ""
    if query and len (query)>0:
        if isinstance(query, (list,tuple)):
            for q in query:
                #q = str(q, 'utf-8')
                for param in config.QUERY_PARAMS:
                    q = replaceStr(sString=q, findStr=param, repStr=config.QUERY_PARAMS[param], ignoreCase=True, addQuotes="'")
                    #if param in q:
                    #    q = q.replace(param, config.QUERY_PARAMS[param])
                    p("glob->setQueryWithParams: replace param %s with value %s" % (str(param), str(config.QUERY_PARAMS[param])), "ii")
                qRet += q
        else:
            #query= str (query, 'utf-8')

            for param in config.QUERY_PARAMS:
                if param in query:
                    query = replaceStr(sString=query, findStr=param, repStr=config.QUERY_PARAMS[param], ignoreCase=True,addQuotes="'")
                    p("glob->setQueryWithParams: replace param %s with value %s ..." % (str(param), str(config.QUERY_PARAMS[param])), "ii")
            qRet += query
    else:
        qRet = query
    return qRet

def replaceStr (sString,findStr, repStr, ignoreCase=True,addQuotes=None):
    if addQuotes and isinstance(repStr,str):
        repStr="%s%s%s" %(addQuotes,repStr,addQuotes)

    if ignoreCase:
        pattern = re.compile(re.escape(findStr), re.IGNORECASE)
        res = pattern.sub (repStr, sString)
    else:
        res = sString.replace (findStr, repStr)
    return res

def decodeStrPython2Or3 (sObj, un=True):
    pVersion = sys.version_info[0]

    if 3 == pVersion:
        return sObj
    else:
        if un:
            return unicode (sObj)
        else:
            return str(sObj).decode("windows-1255")

def setDicConnValue (connJsonVal=None, connType=None, connName=None,
                     connObj=None, connFilter=None, connUrl=None,
                     extraConnVal=None,fileToLoad=None,
                     isSql=False, isTarget=False, isSource=False):
    retVal = {eConnValues.connName:connName,
              eConnValues.connType:connType.lower() if connType else None ,
              eConnValues.connUrl:connUrl,
              eConnValues.connUrlExParams:extraConnVal,
              eConnValues.connObj:connObj,
              eConnValues.connFilter:connFilter,
              eConnValues.fileToLoad:fileToLoad,
              eConnValues.connIsSql:isSql,
              eConnValues.connIsSrc:isSource,
              eConnValues.connIsTar:isTarget}

    if isinstance( connJsonVal, (tuple,list) ):
        if len (connJsonVal) == 1:
            retVal[eConnValues.connName] = connJsonVal[0]
        elif len (connJsonVal) >= 2:
            retVal[eConnValues.connName] = connJsonVal[0]
            retVal[eConnValues.connObj]  = connJsonVal[1]
            if retVal[eConnValues.connType] is None:
                retVal[eConnValues.connType] = connJsonVal[0].lower()
            if len (connJsonVal) == 3:
                retVal[eConnValues.connFilter]= connJsonVal[2]
        else:
            err = "glob->_setDicConnValue: Connection paramter is not valid %s must have 1,2 or 3 params: %s " %(str(connJsonVal))
            p(err, "e")
            raise Exception(err)

        if retVal[ eConnValues.connName ] is None:
            err = "glob->_setDicConnValue: Connection Name is not defined: %s " %(connJsonVal)
            p(err, "e")
            raise Exception(err)

        if retVal[eConnValues.connUrl] is None:
            if retVal[eConnValues.connName] in config.CONN_URL:
                connUrl = config.CONN_URL[ retVal[eConnValues.connName] ]
                retVal[eConnValues.connUrl] = connUrl
                if isinstance(connUrl, (dict, OrderedDict) ):
                    if eConnValues.connType in connUrl:
                        retVal[eConnValues.connType] = connUrl[eConnValues.connType].lower()
                    if eConnValues.connUrl in connUrl:
                        retVal[eConnValues.connUrl] = connUrl[eConnValues.connUrl]
                    if eConnValues.fileToLoad in connUrl:
                        retVal[eConnValues.fileToLoad] = connUrl[eConnValues.fileToLoad]
            else:
                err = "glob->_setDicConnValue: Connection Name %s is not defined in CONN_URL config. define names are : %s  "  %(retVal[eConnValues.connName], str(list(config.CONN_URL.keys())))
                p(err, "e")
                raise Exception(err)
        # remove number from connection type in case we used it in config.CONN_URL
        # sample : sql1 - will be rename to sql as a type
        retVal[eConnValues.connType] = ''.join([i for i in retVal[eConnValues.connType].lower() if not i.isdigit()])

        #################   ACCESS - ADD EXTRA PARAMTERS -> Access File DB
        if eDbType.ACCESS == retVal[eConnValues.connType] :
            if retVal[eConnValues.connUrlExParams] is not None:
                retVal[eConnValues.connUrl] = retVal[eConnValues.connUrl][0] % (retVal[eConnValues.connUrl][1] + str(retVal[eConnValues.connUrlExParams].split(".")[0] + ".accdb"))
            else:
                err = "glob->_setDicConnValue: Connection %s is missing Access file " % (retVal[eConnValues.connName])
                p(err, "e")
                raise Exception(err)

        #################   LOAD SQL QUERIES FROM FILE - ADD EXTRA PARAMTERS -> File loacation
        foundQuery = False
        allParams  = []
        if retVal[eConnValues.fileToLoad] is not None:
            sqlFile = "%s.sql" %retVal[eConnValues.fileToLoad] if retVal[eConnValues.fileToLoad][-4:]!=".sql" else retVal[eConnValues.fileToLoad]
            if os.path.isfile(sqlFile):
                with io.open(sqlFile, 'r', encoding=config.PARSER_FILE_ENCODE) as inp:
                    sqlScript = inp.readlines()
                    allQueries =  queryParsetIntoList (sqlScript, getPython=True, removeContent=True, dicProp=None, pythonWord=config.PARSER_SQL_MAIN_KEY)

                    for q in allQueries:
                        allParams.append(q[1])
                        if q[0] and q[0] == retVal[eConnValues.connObj]:
                            retVal[eConnValues.connObj]     = q[1]
                            retVal[eConnValues.connIsSql]   = True
                            foundQuery = True
                            break
                if not foundQuery:
                    err = "glob->_setDicConnValue: There is paramter %s which is not found in %s, existing keys: %s " % (retVal[eConnValues.connObj],sqlFile, str(allParams))
                    p(err, "e")
                    raise Exception(err)
            else:
                err = "glob->_setDicConnValue: %s is not found  " % (sqlFile)
                p(err, "e")
                raise Exception(err)


        retVal[eConnValues.connType] = isDbType( retVal[eConnValues.connType] )

        if  retVal[eConnValues.connName] is not None and \
            retVal[eConnValues.connType] is not None and \
            retVal[eConnValues.connObj] is not None and \
            retVal[eConnValues.connUrl] is not None:
                p ("glob->_setDicConnValue: Connection params: %s " %(str(retVal)),"ii")
                return retVal

        err = "glob->_setDicConnValue: Connection params are not set: %s " % (str(retVal))
        p(err, "e")
        raise Exception(err)

def getDicKey (etlProp, allProp):
    etlProp = str(etlProp).lower() if etlProp else ''

    if etlProp in ePopEtlProp.dicOfProp:
        etlProps = ePopEtlProp.dicOfProp[ etlProp ]

        filterSet = set (etlProps)
        allSet    = set ([str(x).lower() for x in allProp])
        isExists = filterSet.intersection(allSet)

        if len (isExists) > 0:
            return isExists.pop()
    return None

def filterFiles (modelToExec, dirData=None, includeFiles=None, notIncludeFiles=None ):
    dirData          = dirData if dirData else config.DIR_DATA
    notIncludeFiles = notIncludeFiles if notIncludeFiles else config.FILES_NOT_INCLUDE
    notIncludeFilesL=[x.lower().replace(".json","") for x in notIncludeFiles]
    includeFiles    = includeFiles if includeFiles else config.FILES_INCLUDE

    jsonFiles = [pos_json for pos_json in os.listdir(dirData) if pos_json.endswith('.json')]

    jsonFilesDic    = {x.lower().replace(".json",""):x for x in jsonFiles}



    if  notIncludeFiles:
        notIncludeDict = {x.lower().replace(".json", ""): x for x in notIncludeFiles}
        for f in jsonFilesDic:
            if f in notIncludeDict:
                p('%s: NOT INCLUDE: Folder:%s, file: %s NOT IN USED, REMOVED >>>>' % (modelToExec, str(config.DIR_DATA), f),"ii")
                jsonFiles.remove( jsonFilesDic[f] )
        for f in notIncludeDict:
            if f not in jsonFilesDic:
                p('%s: NOT INCLUDE: Folder: %s, file: %s not exists.. Ignoring>>>>>' % (modelToExec, str(config.DIR_DATA), f), "ii")

    if  includeFiles:
        includeDict = {x.lower().replace(".json", ""): x for x in includeFiles}
        for f in jsonFilesDic:
            if f not in includeDict:
                p('%s: INCLUDE: Folder:%s, file: %s NOT IN USED, REMOVED >>>>'% (modelToExec, str(config.DIR_DATA), f), "ii")
                jsonFiles.remove( jsonFilesDic[f] )
        for f in includeDict:
            if f not in jsonFilesDic:
                p('%s: INCLUDE: Folder: %s, file: %s not exists.. Ignoring >>>>' % (modelToExec, str(config.DIR_DATA), f), "ii")

    return jsonFiles


import io, re

def _removeComments (listQuery, endOfLine='\n'):
    retList = []
    for s in listQuery:
        isTup = False
        if isinstance(s, (tuple, list) ):
            pre = s[0].strip() if s[0] else None
            post= s[1].strip()
            isTup = True
        else:
            post = s.strip()

        post = re.sub(r"--.*$", r"",    post, flags= re.IGNORECASE | re.MULTILINE | re.UNICODE  ).replace ("--","")
        post = re.sub(r'\/\*.*\*\/',    "", post, flags=re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL )
        post = re.sub(r"print .*$", r"",post, flags=re.IGNORECASE | re.MULTILINE | re.UNICODE ).replace ("print ","")

        if endOfLine:
            while len (post)>1 and post[0:1] == "\n":
                post = post[1:]

            while len (post)>1 and post[-1:] == "\n":
                post = post[:-1]

        if not post or len(post) == 0:
            continue
        else:
            if isTup:
                retList.append ( (pre,post,) )
            else:
                retList.append (post)

    return retList

def _getPythonParam (queryList, mWorld="popEtl"):
    ret = []
    for query in queryList:
        # Delete all rows which are not relevant
        # Regex : <!popEtl XXXX/>
        #fPythonNot = re.search(r"<!%s([^>].*)/>" % (mWorld), query,flags=re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.S)
        # Regex : <!popEtl> ......... </!popEtl>
        reg = re.finditer(r"<!%s(.+?)</!%s>" % (mWorld, mWorld), query,flags=re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.S )
        if reg:
            for regRemove in reg:
                query = query.replace (regRemove.group(0),"")

        # Add python queries into return list
        # Regex : <popEtl STRING_NAME> ....... </popEtl>
        #fPython2    = re.search(r"<%s.*/%s>" % (mWorld,mWorld),   query, flags = re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.S)

        # Regex : <popEtl STRING_NAME>......</popEtl> --> Take string to the end
        reg = re.finditer(r"<%s(.+?)>(.+?)</%s>" %(mWorld,mWorld), query,flags=re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.S)

        if reg:
            for i, regFind in enumerate (reg):
                pythonSeq = regFind.group(0)
                pythonVar = regFind.group(1).strip()
                querySql  = regFind.group(2).strip()

                if i == 0 and regFind.start()>0:
                    queryStart = query[: query.find(pythonSeq)].strip()
                    if queryStart and len (queryStart)>0:
                        ret.append((None, queryStart))

                ret.append((pythonVar, querySql))
        else:
            if query and len(query.strip()) > 0:
                ret.append ( (None, query.strip()) )
    return ret

def _getAllQuery (longStr, splitParam = ['GO',u';']):
    sqlList = []
    for splP in splitParam:
        if len(sqlList) == 0:
            sqlList = longStr.split (splP)
        else:
            tmpList = list([])
            for sql in sqlList:
                tmpList.extend (sql.split(splP))
            sqlList = tmpList
    return sqlList

def _replaceStr (sString,findStr, repStr, ignoreCase=True):
    if ignoreCase:
        pattern = re.compile(re.escape(findStr), re.IGNORECASE)
        res = pattern.sub (repStr, sString)
    else:
        res = sString.replace (findStr, repStr)
    return res

def _replaceProp(allQueries, dicProp):
    ret = []
    for query in allQueries:
        if isinstance(query, (list,tuple)):
            pr1 = query[0]
            pr2 = query[1]
        else:
            pr2 = query
        if not pr1 or pr1 and pr1!="~":
            for prop in dicProp:
                pr2= ( _replaceStr(sString=pr2, findStr=prop, repStr=dicProp[prop], ignoreCase=True) )

        tupRet = (pr1, pr2,) if isinstance(query, (list,tuple)) else pr2
        ret.append (tupRet)
    return ret

def queryParsetIntoList (sqlScript, getPython=True, removeContent=True, dicProp=None, pythonWord="popEtl"):
    if isinstance(sqlScript, (tuple,list)):
        sqlScript = "".join(sqlScript)
    # return list of sql (splitted by list of params)
    allQueries = _getAllQuery(longStr=sqlScript, splitParam = ['GO',u';'])

    if getPython:
        allQueries = _getPythonParam(allQueries, mWorld=pythonWord)

    if removeContent:
        allQueries = _removeComments(allQueries)

    if dicProp:
        allQueries = _replaceProp(allQueries, dicProp)

    return allQueries




# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of popEye
#
# popEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# popEye is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cadenceEtl.  If not, see <http://www.gnu.org/licenses/>.

from popEtl.config                  import config
from popEtl.glob.loaderFunctions    import *
from popEtl.glob.glob               import p
from popEtl.glob.enums              import eConnValues, eDbType

from popEtl.connections.db     import cnDb
from popEtl.connections.file   import cnFile


class connector ():
    def __init__ (self, connDic  ):
        # connProp, connUrl=None, isSql=False, fileName=None
        # Expose all paramters into mapper and loader classes
        if connDic is None:
            p ("CONNECTOR->init: %s is Not valid connection .... quiting ...." %(str(connDic)) ,"e")
            return

        self.objClass   = None
        self.cName      = connDic [ eConnValues.connName ]

        className   = eval ( config.CONNECTIONS_ACTIVE[connDic [eConnValues.connType] ] )

        if className:
                self.objClass   = className(connDic)
                self.cType      = self.objClass.cType
                self.cName      = self.objClass.cName
                self.cursor     = self.objClass.cursor
                self.conn       = self.objClass.conn
                self.cColumns   = self.objClass.cColumns
        else:
            p ("CONNECTOR->init: %s is Not valid connection .... quiting ...." %(str(connDic)) ,"e")
            return

    # GENERAL
    def setColumns(self, sttDic):
        p ("CONNECTOR->setColumns: setColumn type:%s, name: %s " %(self.cType, self.cName), "ii")
        return self.objClass.setColumns(sttDic)

    def create (self, stt=None, seq=None,tblName=None):
        if self.objClass:
            objName = "query" if self.objClass.cIsSql else self.cName
            if seq:
                p ("CONNECTOR->create: create type:%s, name: %s WITH SEQUENCE: %s" % (self.cType , objName, str (seq)) , "ii")
            elif tblName:
                p("CONNECTOR->create: create new table type:%s, name: %s " % (self.cType, tblName), "ii")
            else:
                p ("CONNECTOR->create: create type:%s, name: %s " % (self.cType , objName) , "ii")
        self.objClass.create(stt=stt, seq=seq, tblName=tblName)

    def getColumns(self):
        #p("CONNECTOR->getColumns: Get column strucure schema type:%s, name: %s " % (self.cType, self.cName), "ii")
        return self.objClass.getColumns()

    def structure (self,stt ,addSourceColumn=False,tableName=None, sqlQuery=None):
        #p ("CONNECTOR->structure: Get current schema type:%s, name: %s " %(self.cType, self.cName) ,"ii")
        return self.objClass.structure(stt, addSourceColumn=addSourceColumn,tableName=tableName, sqlQuery=sqlQuery)

    def truncate (self):
        #p ("CONNECTOR->truncate: Truncating schema type:%s, name: %s " %(self.cType, self.cName) ,"ii")
        return self.objClass.truncate()

    def execSP (self, sqlQuery):
        #p ("CONNECTOR->execSP: schema:%s, name: %s, executing query %s " %(self.cType, self.cName,sqlQuery) ,"ii")
        return self.objClass.execSP (sqlQuery)

    def transferToTarget(self, dstObj, sttDic):
        #p("CONNECTOR->transferToTarget: Transfer data from %s, type: %s to %s, type: %s " % (self.cName, self.cType, dstObj.cName, dstObj.cType), "ii")
        pp = False if dstObj.cType in [eDbType.FILE] else True
        srcVsTar= []
        fnDic   = {}

        if sttDic:
            for cnt, t in enumerate(sttDic):
                newMap = (sttDic[t]["s"] if "s" in sttDic[t] else "''",t,)
                srcVsTar.append (newMap)
                if "f" in sttDic[t]:
                    fnc = eval(sttDic[t]["f"])
                    fnDic[cnt] = fnc if isinstance(fnc, (list, tuple)) else [fnc]
                # key is tuple of column: (c1,c2, c3 ...)
                # values is (location, string function, toEvel (true,false) )
                elif "c" in sttDic[t] or "ce" in sttDic[t]:
                    colList = sttDic[t]["c"][0]
                    colFun  = sttDic[t]["c"][1]
                    toEval  = True if "ce" in sttDic[t] else False
                    newKey  = []
                    for j, tc in enumerate(sttDic):
                        if tc in colList:
                            newKey.append ( j )
                    fnDic[ tuple(newKey) ] = (cnt, colFun, toEval)
        return self.objClass.transferToTarget(dstObj=dstObj, srcVsTar=srcVsTar, fnDic=fnDic, pp=pp)

    def loadData(self, srcVsTar, results, numOfRows, cntColumn):
        return self.objClass.loadData(srcVsTar, results, numOfRows, cntColumn)

    def sqlTargetMapping(self):
        #p("CONNECTOR->sqlTargetMapping: Update cColumns and cColumnsTDic, type %s, name: %s " % (self.cType,self.cName), "ii")
        return self.objClass.sqlTargetMapping()

    def allStrucure(self, filterDic=None, withType=False):
        #p("CONNECTOR->allStrucure: Reciave all DB object, type %s, name: %s, filter %s, withType: %s  " % (self.cType, self.cName, str(filterDic), str(withType)), "ii")
        return self.objClass.allStrucure(filterDic, withType)


    def dfFromTable (self, columns=None, index_col=None):
        #p("CONNECTOR->dfFromTable: Load into dataframe all data type:%s, name: %s " % (self.cType, self.cName), "ii")
        return self.objClass.dfFromTable (columns, index_col)

    def dfFromQuery (self, sql, index_col=None):
        #p("CONNECTOR->dfFromQuery: SQL Query Load into dataframe all data type:%s, name: %s , query: %s" % (self.cType, self.cName, str(sql)), "ii")
        return self.objClass.dfFromQuery (sql, index_col)

    def minValues (self, colToFilter=None, resolution=None, periods=None, startDate=None):
        #p("CONNECTOR->minValues: Return MINIMUM values of field %s data type:%s, name: %s " % (colToFilter, self.cType, self.cName), "ii")
        return self.objClass.minValues (colToFilter=colToFilter, resolution=resolution, periods=periods, startDate=startDate)

    def merge (self, mergeTable, mergeKeys=None):
        #p("CONNECTOR->merge: type: %s, merge source %s with destination %s, keys: %s " % (self.cType, self.cName, str(mergeTable), str (mergeKeys)), "ii")
        return self.objClass.merge (mergeTable=mergeTable, mergeKeys=mergeKeys)

    def close (self):
        #p ("CONNECTOR->close: CLOSING CONNECTION type:%s, name: %s " %(self.cType, self.cName) ,"ii")
        self.objClass.close()

    def cntRows (self):
        #p ("CONNECTOR->cntRows: Count rows type:%s, name: %s " %(self.cType, self.cName) ,"ii")
        return self.objClass.cntRows()


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of popEye
#
# popEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# popEye is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cadenceEtl.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import  (absolute_import, division, print_function)
__metaclass__ = type

import os
import json
import multiprocessing
import time
import io
from collections import OrderedDict, Counter

from popEtl.config                  import config
from popEtl.glob.glob               import p, setQueryWithParams, getDicKey, setDicConnValue, filterFiles
from popEtl.glob.enums              import eConnValues, ePopEtlProp
from popEtl.connections.dbSqlLite   import sqlLite
from popEtl.connections.connector   import connector
from popEtl.glob.globalDBFunctions  import logsToDb

##############   Not in used - need to check            #####################################
# Saving into file new sequence
def addIncSaveToFile (df, columnInc, srcObj):
    if df is not None and columnInc in df.columns:
        maxValue = df[columnInc].max()
        if str(maxValue).lower() == 'nan':
            p("loader->addSeqSaveToFile: df is empty - there are no rows to add into column %s " % (str(columnInc)), "ii")
            return
        else:
            db = sqlLite(os.path.join(config.DIR_DATA, config.SEQ_DB_FILE_NAME))
            db.execReplaceSql(srcObj.objName, columnInc, maxValue)
            db.close()
    else:
        p ("loader->addSeqSaveToFile: df is not exists or column %s is not in dataframe" %(str(columnInc)) , "ii")

def addIncemenral (inc, isSQL, srcObj, srcMapping=None):
    if not inc:
        return isSQL, None, None

    columnSeq = inc['column']
    columnStart = inc['start'] if 'start' in inc else 0
    db = sqlLite ( os.path.join(config.DIR_DATA, config.SEQ_DB_FILE_NAME) )
    curSeq = db.execSqlGetValue (srcObj.cName , columnSeq)
    if curSeq:
        columnStart = curSeq
    db.close()


    if isSQL:
        isSQL = isSQL.replace (config.QUERY_SEQ_TAG_VALUE, str (columnStart)).replace (config.QUERY_SEQ_TAG_FIELD, str (columnSeq))
    else:
        tblName    = srcObj.objName
        if srcMapping:
            isSQL = "Select "+",".join (srcMapping)+" From "+tblName+" Where "+str(columnSeq)+" > "+str(columnStart)
        else:
            isSQL = "Select " + "*" + " From " + tblName + " Where " + str(columnSeq) + " > " + str(columnStart)

    p("loader->addSeq: Sequence is ON, field %s, start from %s, file is update, sql: %s" % (str(columnSeq),str(columnStart),str(isSQL)), "ii")
    return isSQL, columnSeq, columnStart
##############   Not in used - need to check            #####################################


# Will merge table with connection in source object
def _execMerge (dstDict, mergeConn, sttDic, toCreate=True):

    mergeKeys = None
    if isinstance( mergeConn , (list,tuple) ):
        mergeTable = mergeConn[0]
        mergeKeys  = mergeConn[1]
        if len (mergeConn) == 3:
            toCreate = mergeConn[2]
    else:
        mergeTable = mergeConn

    dstObj = connector(connDic=dstDict)
    # create target table same as source one
    if toCreate:
        dstObj.create (stt=sttDic, tblName=mergeTable )

    dstObj.merge (mergeTable=mergeTable, mergeKeys=mergeKeys)
    dstObj.close()

def _appendPartitions (srcDic, partition):
    srcName = srcDic [eConnValues.connObj]
    ret = []
    sqlStart = "SELECT * FROM %s WHERE " %(srcName)
    if eConnValues.partitionCol in partition:
        colToFilter = partition[ eConnValues.partitionCol ]
    else:
        p("loader->_appendPartitions: There is partition without column definition table: %s ..." % (srcName), "e")
        return ret

    if eConnValues.partitionAgg in partition:
        resolution = partition[ eConnValues.partitionAgg ]
        srcObj = connector(connDic=srcDic)

        # Set starting from partition
        if eConnValues.partitionStart in partition:
            minDate = srcObj.minValues (resolution=resolution, periods=partition[ eConnValues.partitionStart ])
        else:
            minDate = srcObj.minValues (colToFilter=colToFilter, resolution=resolution)

        lastDate = srcObj.minValues (resolution=resolution, periods=0)

        while minDate<lastDate:
            newDate = srcObj.minValues (resolution=resolution, periods=1, startDate=minDate)
            if newDate < lastDate:
                sqlWhere = sqlStart+"%s >= '%s' and %s < '%s'" %(colToFilter,minDate,colToFilter,newDate)
            else:
                sqlWhere = sqlStart + "%s >= '%s'" % (colToFilter, minDate)
            minDate = newDate
            ret.append(sqlWhere)
    else:
        p ("loader->_appendPartitions: There is partition without aggragation function, needs to have 'agg' with values of 'd', 'm' or 'y', table: %s ..." %(str (src)) , "e")
    return ret

# jMap, src, dst
def _execTarget (dstDict):
    if dstDict[ eConnValues.connFilter ]:
        dstObj = connector( connDic=dstDict)
        sql = "Delete From " + dstObj.cName + " Where " + dstDict[ eConnValues.connFilter ]
        sql = setQueryWithParams(sql)
        p("loader->_execTarget: Destination %s have delete query %s, deleting target " % (dstObj.cName, sql), "ii")
        dstObj.execSP(sql)
        dstObj.close()
        return

    if config.TO_TRUNCATE:
        dstObj = connector(connDic=dstDict)
        p("loader->_execTarget: Destination %s is trancating  " % (dstObj.cName), "ii")
        dstObj.truncate()
        dstObj.close()
        return
    return

def _execSql (connDict):
    if connDict[ eConnValues.connIsSql ] == True:
        connObj = connector(connDic=connDict)
        sql = setQueryWithParams( connDict[ eConnValues.connObj ] )
        p("loader->_execSql: exec  %s " % (sql), "i")
        connObj.execSP(sql)
        connObj.close()
        return



def _updateSourceTargetCompareLog (js):
    isRepoTblsExists  = False
    connUrl     = None
    connType    = None

    with open(os.path.join(config.DIR_DATA, js)) as jsonFile:
        jText = json.load(jsonFile, object_pairs_hook=OrderedDict)

    p("loader->_updateSourceTargetCompareLog: Loading from file %s COMPARE number of rows in Source vs target " % (str(js)),"i")
    for connType in config.CONN_URL:
        if "repo" in connType.lower():
            connUrl     = config.CONN_URL[connType]
            connType    = connType.lower()
            connType    = connType.replace("repo","")
            isRepoTblsExists = True
            break

    if isRepoTblsExists:
        for jMap in jText:
            keys        = map(lambda x: x.lower(), jMap.keys())
            dst         = {'target', 'tar'}.intersection(set(keys))
            src         = {'source','src'}.intersection(set(keys))
            query       = {'query'}.intersection(set(keys))

            # exists source/query AND destination OR destination and merge only
            if len(dst)>0 and (len(src)>0 or len(query)>0)  :
                query   = query.pop()           if len (query)>0    else None
                isSql   = True                  if query            else False

                # will use dst if exists. if not-> will use source if exists, if not will use query                                 dst->src->query
                dst     = dst.pop() if len(dst)>0 else src.pop() if len(src)>0 else query
                src     = src.pop() if len(src)>0 else query if query else dst
                src = jMap[src]
                dst = jMap[dst]



                if "access" in src[0]:
                    accessFilePath = config.CONN_URL["access"][0] % (
                    config.CONN_URL["access"][1] + str(js.split(".")[0] + ".accdb"))
                    srcObj = connector(connProp=src, connUrl=accessFilePath, isSql=isSql)
                else:
                    srcObj = connector(connProp=src, isSql=isSql)

                dstObj = connector( dst )
                tblDstName = dst[1]
                tblDstType = dst[0]
                tblSrcName = src[1]
                tblSrcType = src[0]
                cntSrc = srcObj.cntRows()
                cntTar = dstObj.cntRows()
                srcObj.close()
                dstObj.close()

                localTime = time.localtime()
                timeStr = time.strftime("%m/%d/%Y %H:%M:%S", localTime)

                logObj = connector ([connType,"logs"], connUrl=connUrl)
                sql = "Insert into "+config.LOGS_COUNT_SRC_DST+" select "
                sql+="'" + timeStr + "'"
                sql += "'" + tblDstName + "',"
                sql += "'" + tblDstType + "',"
                sql +=  str(cntTar)  + ","
                sql += "'" + tblSrcName + "',"
                sql += "'" + tblSrcType + "',"
                sql += str(cntSrc) + ""

                logObj.cursor.execute (sql)
                logObj.conn.commit()
                logObj.close()

# jMap, src, dst, sttDic, isSQL, merge, inc, seq
def _execLoading ( params ):
    (srcDict, dstDict, mergeConn, sttDic, jFileName, cProc, tProc) = params
    if srcDict and dstDict:
        p("loader->_execLoading: loading %s out of %s, src: %s, dst: %s " %(str(cProc), str(tProc), str(srcDict[eConnValues.connName]), str(dstDict[eConnValues.connName])), "i")

        # Managing Destination table
        _execTarget(dstDict=dstDict)

        # True / False indication
        addSourceColumn = False
        if sttDic and config.STT_INTERNAL in sttDic:
            addSourceColumn = sttDic[config.STT_INTERNAL]
            del sttDic[config.STT_INTERNAL]

        srcObj = connector(connDic=srcDict)
        dstObj = connector(connDic=dstDict)

        # Check if source is same as target connection (only for merge option)
        if  srcDict[eConnValues.connType] == dstDict[eConnValues.connType] and \
            srcDict[eConnValues.connObj] == dstDict[eConnValues.connObj]:
            p('loader->execLoading: TYPE: %s, SOURCE and TARGET %s object are identical.. will check if there is merge >>>>>' % (str(srcDict[eConnValues.connType]), str(srcDict[eConnValues.connObj])), "ii")

        else:
            sttDic = srcObj.structure(stt=sttDic, addSourceColumn=addSourceColumn)
            # isSQL, columnInc, columnStart = addIncemenral (inc, isSQL, srcObj, srcMapping)
            srcObj.transferToTarget(dstObj=dstObj, sttDic=sttDic)
            srcObj.close()
            dstObj.close()

    if mergeConn:
        p("loader->_execLoading: MERGE !!!! " , "i")
        if not dstDict:
            dstDict = srcDict
        _execMerge(dstDict=dstDict, mergeConn=mergeConn, sttDic=None, toCreate=True)
        return

    if config.LOGS_IN_DB : logsToDb( str(jFileName)+":"+str(dstDict[eConnValues.connName]) )

def _extractNodes (jText,jFileName,sourceList=None, destList=None, singleProcess=None):
    processList = []
    loadedObject= []
    cProc       = 0
    for jMap in jText:
        # if there is a list - will exec nodes in on a single process
        if isinstance(jMap, (list,tuple)):
            loadedObject = _extractNodes(jText=jMap, jFileName=jFileName,
                                         sourceList=sourceList, destList=destList,
                                         singleProcess=True)
            continue


        toLoad = True
        sttDic = None
        keys = [x.lower() for x in jMap.keys()]
        locSql          = None
        sourceConn      = getDicKey(ePopEtlProp.src, keys)
        queryConn       = getDicKey(ePopEtlProp.qry, keys)
        targetConn      = getDicKey(ePopEtlProp.tar, keys)
        mergeConn       = getDicKey(ePopEtlProp.mrg, keys)
        seq             = getDicKey(ePopEtlProp.seq, keys)
        stt             = getDicKey(ePopEtlProp.stt, keys)
        targetMapping   = getDicKey(ePopEtlProp.map, keys)

        partition       = getDicKey(ePopEtlProp.par,keys)
        inc             = getDicKey(ePopEtlProp.inc,keys)
        execSql         = getDicKey(ePopEtlProp.exe,keys)


        sttDic          = jMap[stt]  if stt and len (jMap[stt])>0 else None
        mergeConn       = jMap[mergeConn]  if mergeConn and len (jMap[mergeConn])>0 else None

        srcDic = setDicConnValue(connJsonVal=jMap[sourceConn], extraConnVal=jFileName, isSource=True) if sourceConn else None
        if queryConn:
            if srcDic:
                p("loader->_extractNodes: Found %s and %s, will use %s as source data " %(ePopEtlProp.src, ePopEtlProp.qry,ePopEtlProp.qry),"i" )
            srcDic = setDicConnValue(connJsonVal=jMap[queryConn], extraConnVal=jFileName, isSource=True, isSql=True) if queryConn else None

        tarDic = setDicConnValue(connJsonVal=jMap[targetConn], extraConnVal=jFileName, isTarget=True) if targetConn else None

        execDic= setDicConnValue(connJsonVal=jMap[execSql], extraConnVal=jFileName, isSql=True) if execSql else None
        if execDic:
            locSql = keys.index(execSql)
            if locSql==0: _execSql(execDic)
            if len(keys) == 1:
                continue


        # if there is source and target or merge with source/target
        if (srcDic and tarDic) or (mergeConn and (srcDic or tarDic)):
            # update sttDic with mapping -> if exists
            if tarDic:
                if sttDic:
                    sttDicTemp = sttDic
                    sttDic = OrderedDict()
                    if targetMapping:
                        for t in targetMapping:
                            if t in sttDicTemp:
                                sttDic[t] = sttDicTemp[t]
                                if "s" not in sttDic[t]:
                                    sttDic[t]["s"] = targetMapping[t]
                            else:
                                sttDic[t] = {"s": targetMapping[t]}
                    if sttDicTemp:
                        for t in sttDicTemp:
                            if t not in sttDic: sttDic[t] = sttDicTemp[t]
                elif targetMapping:
                    sttDic = OrderedDict()
                    for t in targetMapping:
                        sttDic[t] = {"s": targetMapping[t]}

            sttDic = sttDic if sttDic and len(sttDic) > 0 else None
            # Update sttDic
            if sttDic:
                sttDic[config.STT_INTERNAL] = True if ePopEtlProp.sttA == stt.lower() and not targetMapping else False

            if sourceList:
                sourceList = [x.lower() for x in sourceList]
                toLoad = True if srcDic[eConnValues.connName] in sourceList or srcDic[eConnValues.connObj] in sourceList else False
            if destList:
                destList = [x.lower() for x in destList]
                toLoad = True if tarDic[eConnValues.connName] in destList or tarDic[eConnValues.connObj] in destList else False

            if toLoad:
                if srcDic and tarDic:
                    p('loader->_extractNodes: SOURCE %s -> TARET %s ; %s -->  %s .......' % (str(srcDic[ eConnValues.connType ]), str(tarDic[ eConnValues.connType ]), srcDic[ eConnValues.connName ], tarDic[ eConnValues.connName ]), "ii")
                if tarDic and mergeConn:
                    p('loader->_extractNodes: Type: %s, TAREGT -> MERGE %s  .......' % (str(tarDic[eConnValues.connType]), tarDic[eConnValues.connName]), "ii")
                if srcDic and mergeConn:
                    p('loader->_extractNodes: Type: %s, SOURCE -> MERGE %s  .......' % (str(srcDic[eConnValues.connType]), srcDic[eConnValues.connName]), "ii")
                # if partition --> change to all partitions
                # update list of data to process:
                if partition:
                    if inc:
                        p('loader->_extractNodes: Cannot have incremental and partiton loading methods.. will use partiton method >>>>>',"ii")
                    if not queryConn or len(queryConn) < 1:
                        newSqlList = _appendPartitions(srcDic, partition)
                        for newSql in newSqlList:
                            cProc += 1
                            tmpParDic = setDicConnValue (connJsonVal=None,
                                                         connType=srcDic [ eConnValues.connType ],
                                                         connName=srcDic [ eConnValues.connName ],
                                                         connObj=newSql,
                                                         connFilter=None,
                                                         connUrl=None, extraConnVal=None, isSql=True,
                                                         isTarget=False, isSource=True)
                            if singleProcess:
                                _execLoading((tmpParDic, tarDic, mergeConn, sttDic, jFileName, cProc, 1))
                            else:
                                processList.append((tmpParDic, tarDic, mergeConn, sttDic, jFileName, cProc))

                        loadedObject.append("P: %s; " %tarDic[ eConnValues.connObj ])
                    else:
                        cProc += 1
                        p('loader->_extractNodes: Cannot have partition with query as source.. will use query as is, sql: %s >>>>>' % (str(srcDic [ eConnValues.connObj ])), "ii")
                        if singleProcess:
                            _execLoading( (srcDic, tarDic, mergeConn, sttDic, jFileName, cProc, 1) )
                        else:
                            processList.append((srcDic, tarDic, mergeConn, sttDic, jFileName, cProc))
                else:
                    cProc += 1
                    if singleProcess:
                        _execLoading((srcDic, tarDic, mergeConn, sttDic, jFileName, cProc , 1))
                    else:
                        processList.append((srcDic, tarDic, mergeConn, sttDic, jFileName, cProc))
                if tarDic:
                    loadedObject.append("%s; " %(tarDic [ eConnValues.connObj ]))
                if mergeConn:
                    mergTbl = mergeConn[0] if isinstance(mergeConn, (list,tuple)) else mergeConn
                    loadedObject.append("MERGE: %s; " % (mergTbl))
        else:
            p("loader->_extractNodes: There is nothing to do >>>>>>>>>>>>>>", "i")

        # Strat runing all processes per file

    numOfProcesses = len(processList) if len(processList) < config.NUM_OF_PROCESSES else config.NUM_OF_PROCESSES

    # Add total processes to execute
    for i, itemP in enumerate(processList):
        processList[i] = itemP + (cProc,)


    if numOfProcesses > 1:
        proc = multiprocessing.Pool(config.NUM_OF_PROCESSES).map(_execLoading, processList)
    elif numOfProcesses > 0:
            for etl in processList:
                _execLoading(etl)

    return loadedObject

def trasnfer (dicObj=None, sourceList=None, destList=None):
    loadedObject = []
    if dicObj:
        dicObj = list (dicObj) if isinstance(dicObj, (dict,OrderedDict)) else dicObj
        p('loader->loading: loading from Dictionary >>>>>' , "ii")
        loadedObject = _extractNodes(jText=dicObj, jFileName='', sourceList=sourceList, destList=destList)
    else:
        jsonFiles = filterFiles(modelToExec="loader->loading", dirData=None, includeFiles=None, notIncludeFiles=None)

        for index, js in enumerate(jsonFiles):
            with io.open(os.path.join(config.DIR_DATA, js), encoding="utf-8") as jsonFile:           #
                jText = json.load(jsonFile, object_pairs_hook=OrderedDict)

            p("loader->loading: Start loading from file %s, folder: %s >>>>> >>>>>>" %(str(js), str(config.DIR_DATA)), "i")
            loadedObject = _extractNodes(jText, jFileName=js, sourceList=sourceList, destList=destList)
            p("loader->loading: Finish loading from file %s >>>>>>" %(str(js)), "i")
            if config.LOGS_COUNT_SRC_DST: _updateSourceTargetCompareLog(js)

    if config.LOGS_IN_DB: logsToDb()
    p("loader->loading: FINISH LOADING, Loader into : " + str (loadedObject), "i")

if __name__ == '__main__':
    multiprocessing.freeze_support()


from __future__ import print_function

import json
import os
import os.path
import re
import sys
import warnings

from collections import defaultdict
from distutils.command.build_scripts import build_scripts as BuildScripts
from distutils.command.sdist import sdist as SDist

try:
    from setuptools import setup, find_packages
    from setuptools.command.build_py import build_py as BuildPy
    from setuptools.command.install_lib import install_lib as InstallLib
    from setuptools.command.install_scripts import install_scripts as InstallScripts
except ImportError:
    print("popEtl now needs setuptools in order to build. Install it using"
          " your package manager (usually python-setuptools) or via pip (pip"
          " install setuptools).", file=sys.stderr)
    sys.exit(1)

sys.path.insert(0, os.path.abspath('lib'))
from popEtl.release import __version__, __author__

def read_file(file_name):
    """Read file and return its contents."""
    with open(file_name, 'r') as f:
        return f.read()

def read_requirements(file_name):
    """Read requirements file as a list."""
    reqs = read_file(file_name).splitlines()
    if not reqs:
        raise RuntimeError(
            "Unable to read requirements from the %s file"
            "That indicates this copy of the source code is incomplete."
            % file_name
        )
    return reqs

def get_dynamic_setup_params():
    """Add dynamically calculated setup params to static ones."""
    return {
        # Retrieve the long description from the README
        'long_description': read_file('./README.rst'),
        'install_requires': read_requirements('./requirements.txt')
    }

# packages=['', 'glob', 'mapp', 'loader', 'connections'],
#download_url = 'https://github.com/biskilled/popEye-Etl/archive/1.1.0.tar.gz',  # I explain this later on
#install_requires = ['pyodbc', 'pymysql', 'vertica_python', 'cx_Oracle', 'pymongo', 'sqlparse', 'clr', 'pandas'],

static_setup_params = dict(
    name        = 'popEtl',  # How you named your package folder (MyLib)
    version     = __version__,
    description = 'Data modeling managing and transforming data',  # Give a short description about your library
    author      = __author__,
    author_email= 'Tal@BiSkilled.com',  # Type in your E-Mail
    keywords    = ['ETL', 'Data modeling', 'Python', 'Integration', 'Mapping'],   # Keywords that define your package best
    url         = 'https://github.com/biskilled/popEtl',  # Provide either the link to your github or to your website

    project_urls={
        'Bug Tracker'    : 'https://github.com/ansible/ansible/issues',
        'CI: Shippable'  : 'https://app.shippable.com/github/biskilled/popEtl',
        'Code of Conduct': '',
        'Documentation'  : 'https://readthedocs.org/projects/popeye-etl/',
        'Mailing lists'  : '',
        'Source Code'    : 'https://github.com/biskilled/popEtl',
    },
    license='GPLv3+',
    python_requires= '>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    package_dir    = {'': 'lib'},
    packages       = find_packages('lib'),
    package_data   ={
        '': [ ],
    },

    classifiers=[
        'Development Status :: 4 - Beta',       # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    scripts=[],
    data_files=[],
    # Installing as zip files would break due to references to __file__
    zip_safe=False
)


def main():
    """Invoke installation process using setuptools."""
    setup_params = dict(static_setup_params, **get_dynamic_setup_params())
    ignore_warning_regex = (
        r"Unknown distribution option: '(project_urls|python_requires)'"
    )
    warnings.filterwarnings(
        'ignore',
        message=ignore_warning_regex,
        category=UserWarning,
        module='distutils.dist',
    )
    setup(**setup_params)
    warnings.resetwarnings()

if __name__ == '__main__':
    main()

